# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Import Data Providers
"""

from __future__ import unicode_literals

import datetime

from rattail.core import Object
from rattail.db.cache import cache_model


class DataProvider(Object):
    """
    Base class for import data providers.
    """
    importer_class = None
    normalize_progress_message = "Normalizing source data"
    progress = None

    def __init__(self, config, session, importer=None, importer_kwargs={}, **kwargs):
        self.config = config
        self.session = session
        if importer is None:
            self.importer = self.importer_class(config, session, **importer_kwargs)
        else:
            self.importer = importer
        super(DataProvider, self).__init__(**kwargs)
        self.importer.provider = self

    @property
    def model_class(self):
        return self.importer.model_class

    @property
    def key(self):
        """
        Key by which records should be matched between the source data and
        Rattail.
        """
        raise NotImplementedError("Please define the `key` for your data provider.")

    @property
    def model_name(self):
        return self.model_class.__name__

    def cache_model(self, model_class, key, **kwargs):
        """
        Convenience method for caching a model.
        """
        kwargs.setdefault('progress', self.progress)
        return cache_model(self.session, model_class, key=key, **kwargs)

    def setup(self):
        """
        Perform any setup necessary, e.g. cache lookups for existing data.
        """

    def get_data(self, progress=None, normalize_progress_message=None):
        """
        Return the full set of normalized data which is to be imported.
        """
        self.now = datetime.datetime.utcnow()
        self.progress = progress
        self.setup()
        source_data = self.get_source_data(progress=progress)
        data = self.normalize_source_data(source_data, progress=progress)
        self.teardown()
        return data

    def teardown(self):
        """
        Perform any cleanup necessary, after the main data run.
        """

    def get_source_data(self, progress=None):
        """
        Return the data which is to be imported.
        """
        return []

    def normalize_source_data(self, source_data, progress=None):
        """
        Return a normalized version of the full set of source data.
        """
        data = []
        count = len(source_data)
        if count == 0:
            return data
        prog = None
        if progress:
            prog = progress(self.normalize_progress_message, count)
        for i, record in enumerate(source_data, 1):
            record = self.normalize(record)
            if record:
                data.append(record)
            if prog:
                prog.update(i)
        if prog:
            prog.destroy()
        return data

    def normalize(self, data):
        """
        Normalize a source data record.  Generally this is where the provider
        may massage the record in any way necessary, so that its values are
        more "native" and can be used for direct comparison with, and
        assignment to, the target model instance.

        Note that if you override this, your method must return the data to be
        imported.  If your method returns ``None`` then that particular record
        would be skipped and not imported.
        """
        return data

    def int_(self, value):
        """
        Coerce ``value`` to an integer, or return ``None`` if that can't be
        done cleanly.
        """
        try:
            return int(value)
        except (TypeError, ValueError):
            return None


class QueryDataProxy(object):
    """
    Simple proxy to wrap a SQLAlchemy query and make it sort of behave like a
    normal sequence, as much as needed to make a ``DataProvider`` happy.
    """

    def __init__(self, query):
        self.query = query

    def __len__(self):
        return self.query.count()

    def __iter__(self):
        return iter(self.query)


class QueryProvider(DataProvider):
    """
    Data provider whose data source is a SQLAlchemy query.  Note that this
    needn't be a Rattail database query; any database will work as long as a
    SQLAlchemy query is behind it.
    """

    def query(self):
        """
        Return the query which will define the data set.
        """
        raise NotImplementedError

    def get_source_data(self, progress=None):
        """
        Return the data which is to be imported.
        """
        return QueryDataProxy(self.query())



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import warnings
from unittest import TestCase

from sqlalchemy import pool

from rattail.db import config as dbconfig
from rattail.config import RattailConfig


class TestEngineFromConfigFunc(TestCase):

    def test_default_key(self):
        cfg = {'sqlalchemy.url': 'sqlite://'}
        engine = dbconfig.engine_from_config(cfg)
        self.assertEqual(unicode(engine.url), 'sqlite://')

    def test_alternate_key(self):
        cfg = {'other.url': 'sqlite://'}
        engine = dbconfig.engine_from_config(cfg, prefix='other.')
        self.assertEqual(unicode(engine.url), 'sqlite://')

    def test_poolclass(self):

        # default is SingletonThreadPool
        cfg = {'sqlalchemy.url': 'sqlite://'}
        engine = dbconfig.engine_from_config(cfg)
        self.assertEqual(unicode(engine.url), 'sqlite://')
        self.assertIsInstance(engine.pool, pool.SingletonThreadPool)

        # usually we override with NullPool
        cfg = {'sqlalchemy.url': 'sqlite://',
               'sqlalchemy.poolclass': 'sqlalchemy.pool:NullPool'}
        engine = dbconfig.engine_from_config(cfg)
        self.assertEqual(unicode(engine.url), 'sqlite://')
        self.assertIsInstance(engine.pool, pool.NullPool)

    def test_record_changes(self):

        # no attribute is set by default
        cfg = {'sqlalchemy.url': 'sqlite://'}
        engine = dbconfig.engine_from_config(cfg)
        self.assertRaises(AttributeError, getattr, engine, 'rattail_record_changes')

        # usually we override with NullPool
        cfg = {'sqlalchemy.url': 'sqlite://',
               'sqlalchemy.record_changes': 'true'}
        engine = dbconfig.engine_from_config(cfg)
        self.assertTrue(engine.rattail_record_changes)


class TestGetEngines(TestCase):

    def setUp(self):
        self.config = RattailConfig()

    def test_default_section_is_rattail_db(self):
        self.config.set(u'rattail.db', u'keys', u'default')
        self.config.set(u'rattail.db', u'default.url', u'sqlite://')
        engines = dbconfig.get_engines(self.config)
        self.assertEqual(len(engines), 1)
        self.assertEqual(engines.keys()[0], u'default')
        self.assertEqual(unicode(engines[u'default'].url), u'sqlite://')

    def test_custom_section_is_honored(self):
        self.config.set(u'mycustomdb', u'keys', u'default')
        self.config.set(u'mycustomdb', u'default.url', u'sqlite://')
        engines = dbconfig.get_engines(self.config, section=u'mycustomdb')
        self.assertEqual(len(engines), 1)
        self.assertEqual(engines.keys()[0], u'default')
        self.assertEqual(unicode(engines[u'default'].url), u'sqlite://')

    def test_default_prefix_does_not_require_keys_declaration(self):
        self.config.set(u'rattail.db', u'default.url', u'sqlite://')
        engines = dbconfig.get_engines(self.config)
        self.assertEqual(len(engines), 1)
        self.assertEqual(engines.keys()[0], u'default')
        self.assertEqual(unicode(engines[u'default'].url), u'sqlite://')

    def test_default_prefix_falls_back_to_sqlalchemy(self):
        # Still no need to define "keys" option here.
        self.config.set(u'rattail.db', u'sqlalchemy.url', u'sqlite://')
        engines = dbconfig.get_engines(self.config)
        self.assertEqual(len(engines), 1)
        self.assertEqual(engines.keys()[0], u'default')
        self.assertEqual(unicode(engines[u'default'].url), u'sqlite://')

    def test_defined_keys_are_included_in_engines_result(self):
        # Note there is no "default" key here.
        self.config.set(u'rattail.db', u'keys', u'host, store')
        self.config.set(u'rattail.db', u'host.url', u'sqlite:///rattail.host.sqlite')
        self.config.set(u'rattail.db', u'store.url', u'sqlite:///rattail.store.sqlite')
        engines = dbconfig.get_engines(self.config)
        self.assertEqual(len(engines), 2)
        self.assertEqual(sorted(engines.keys()), [u'host', u'store'])
        self.assertEqual(unicode(engines[u'host'].url), u'sqlite:///rattail.host.sqlite')
        self.assertEqual(unicode(engines[u'store'].url), u'sqlite:///rattail.store.sqlite')


class TestGetDefaultEngine(TestCase):

    def setUp(self):
        self.config = RattailConfig()

    def test_default_engine_is_loaded_from_rattail_db_section_by_default(self):
        self.config.set(u'rattail.db', u'keys', u'default')
        self.config.set(u'rattail.db', u'default.url', u'sqlite://')
        engine = dbconfig.get_default_engine(self.config)
        self.assertEqual(unicode(engine.url), u'sqlite://')

    def test_default_engine_is_loaded_from_custom_section_if_specified(self):
        self.config.set(u'mycustomdb', u'keys', u'default')
        self.config.set(u'mycustomdb', u'default.url', u'sqlite://')
        engine = dbconfig.get_default_engine(self.config, section=u'mycustomdb')
        self.assertEqual(unicode(engine.url), u'sqlite://')

    def test_no_engine_is_returned_if_none_is_defined(self):
        engine = dbconfig.get_default_engine(self.config)
        self.assertTrue(engine is None)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

from unittest import TestCase
from mock import patch, DEFAULT, Mock, MagicMock

import sqlalchemy as sa
from sqlalchemy.exc import IntegrityError

from rattail.db import model
from rattail.db.types import GPCType
from rattail.db.changes import record_changes
from rattail.tests import DataTestCase


class SAErrorHelper(object):

    def integrity_or_flush_error(self):
        try:
            from sqlalchemy.exc import FlushError
        except ImportError:
            return IntegrityError
        else:
            return (IntegrityError, FlushError)


class TestPerson(DataTestCase):

    def test_default_display_name_is_generated_from_first_and_last_name_if_both_provided(self):
        person = model.Person(first_name='Fred', last_name='Flintstone')
        self.assertTrue(person.display_name is None)
        self.session.add(person)
        self.session.flush()
        self.assertEqual(person.display_name, 'Fred Flintstone')

    def test_default_display_name_is_generated_from_first_name_if_no_last_name_provided(self):
        person = model.Person(first_name='Fred')
        self.assertTrue(person.display_name is None)
        self.session.add(person)
        self.session.flush()
        self.assertEqual(person.display_name, 'Fred')

    def test_default_display_name_is_generated_from_last_name_if_no_first_name_provided(self):
        person = model.Person(last_name='Flintstone')
        self.assertTrue(person.display_name is None)
        self.session.add(person)
        self.session.flush()
        self.assertEqual(person.display_name, 'Flintstone')


class TestBatch(TestCase):

    # @patch('rattail.db.model.object_session')
    # def test_rowclass(self, object_session):
    #     object_session.return_value = object_session

    #     # no row classes to start with
    #     self.assertEqual(model.Batch._rowclasses, {})

    #     # basic, empty row class
    #     batch = model.Batch(uuid='some_uuid')
    #     batch.get_sqlalchemy_type = Mock(return_value='some_type')
    #     batch.columns = MagicMock()
    #     rowclass = batch.rowclass
    #     self.assertTrue(issubclass(rowclass, model.BatchRow))
    #     self.assertEqual(model.Batch._rowclasses.keys(), ['some_uuid'])
    #     self.assertTrue(model.Batch._rowclasses['some_uuid'] is rowclass)
    #     self.assertFalse(object_session.flush.called)

    #     # make sure rowclass.batch works
    #     object_session.query.return_value.get.return_value = batch
    #     self.assertTrue(rowclass().batch is batch)
    #     object_session.query.return_value.get.assert_called_once_with('some_uuid')

    #     # row class with generated uuid and some columns
    #     batch = model.Batch(uuid=None)
    #     batch.columns = [model.BatchColumn(name='F01'), model.BatchColumn(name='F02')]
    #     model.Batch.get_sqlalchemy_type = Mock(return_value=String(length=20))
    #     def set_uuid():
    #         batch.uuid = 'fresh_uuid'
    #     object_session.flush.side_effect = set_uuid
    #     rowclass = batch.rowclass
    #     object_session.flush.assert_called_once_with()
    #     self.assertEqual(sorted(model.Batch._rowclasses.keys()), sorted(['some_uuid', 'fresh_uuid']))
    #     self.assertTrue(model.Batch._rowclasses['fresh_uuid'] is rowclass)

    def test_get_sqlalchemy_type(self):

        # gpc
        self.assertTrue(isinstance(model.Batch.get_sqlalchemy_type('GPC(14)'), GPCType))

        # boolean
        self.assertTrue(isinstance(model.Batch.get_sqlalchemy_type('FLAG(1)'), sa.Boolean))

        # string
        type_ = model.Batch.get_sqlalchemy_type('CHAR(20)')
        self.assertTrue(isinstance(type_, sa.String))
        self.assertEqual(type_.length, 20)

        # numeric
        type_ = model.Batch.get_sqlalchemy_type('NUMBER(9,3)')
        self.assertTrue(isinstance(type_, sa.Numeric))
        self.assertEqual(type_.precision, 9)
        self.assertEqual(type_.scale, 3)

        # invalid
        self.assertRaises(AssertionError, model.Batch.get_sqlalchemy_type, 'CHAR(9,3)')
        self.assertRaises(AssertionError, model.Batch.get_sqlalchemy_type, 'OMGWTFBBQ')


class TestCustomerPerson(DataTestCase, SAErrorHelper):

    def test_repr(self):
        assoc = model.CustomerPerson(uuid='whatever')
        self.assertEqual(repr(assoc), "CustomerPerson(uuid=u'whatever')")

    def test_customer_required(self):
        assoc = model.CustomerPerson(person=model.Person())
        self.session.add(assoc)
        self.assertRaises(self.integrity_or_flush_error(), self.session.commit)
        self.session.rollback()
        self.assertEqual(self.session.query(model.CustomerPerson).count(), 0)
        assoc.customer = model.Customer()
        self.session.add(assoc)
        self.session.commit()
        self.assertEqual(self.session.query(model.CustomerPerson).count(), 1)

    def test_person_required(self):
        assoc = model.CustomerPerson(customer=model.Customer())
        self.session.add(assoc)
        self.assertRaises(IntegrityError, self.session.commit)
        self.session.rollback()
        self.assertEqual(self.session.query(model.CustomerPerson).count(), 0)
        assoc.person = model.Person()
        self.session.add(assoc)
        self.session.commit()
        self.assertEqual(self.session.query(model.CustomerPerson).count(), 1)

    def test_ordinal_autoincrement(self):
        customer = model.Customer()
        self.session.add(customer)
        assoc = model.CustomerPerson(person=model.Person())
        customer._people.append(assoc)
        self.session.commit()
        self.assertEqual(assoc.ordinal, 1)
        assoc = model.CustomerPerson(person=model.Person())
        customer._people.append(assoc)
        self.session.commit()
        self.assertEqual(assoc.ordinal, 2)


class TestCustomerGroupAssignment(DataTestCase, SAErrorHelper):

    def test_repr(self):
        assignment = model.CustomerGroupAssignment(uuid='whatever')
        self.assertEqual(repr(assignment), "CustomerGroupAssignment(uuid=u'whatever')")

    def test_customer_required(self):
        assignment = model.CustomerGroupAssignment(group=model.CustomerGroup())
        self.session.add(assignment)
        self.assertRaises(self.integrity_or_flush_error(), self.session.commit)
        self.session.rollback()
        self.assertEqual(self.session.query(model.CustomerGroupAssignment).count(), 0)
        assignment.customer = model.Customer()
        self.session.add(assignment)
        self.session.commit()
        self.assertEqual(self.session.query(model.CustomerGroupAssignment).count(), 1)

    def test_group_required(self):
        assignment = model.CustomerGroupAssignment(customer=model.Customer())
        self.session.add(assignment)
        self.assertRaises(IntegrityError, self.session.commit)
        self.session.rollback()
        self.assertEqual(self.session.query(model.CustomerGroupAssignment).count(), 0)
        assignment.group = model.CustomerGroup()
        self.session.add(assignment)
        self.session.commit()
        self.assertEqual(self.session.query(model.CustomerGroupAssignment).count(), 1)

    def test_ordinal_autoincrement(self):
        customer = model.Customer()
        self.session.add(customer)
        assignment = model.CustomerGroupAssignment(group=model.CustomerGroup())
        customer._groups.append(assignment)
        self.session.commit()
        self.assertEqual(assignment.ordinal, 1)
        assignment = model.CustomerGroupAssignment(group=model.CustomerGroup())
        customer._groups.append(assignment)
        self.session.commit()
        self.assertEqual(assignment.ordinal, 2)


# class TestCustomerEmailAddress(DataTestCase):

#     def test_pop(self):
#         customer = model.Customer()
#         customer.add_email_address('fred.home@mailinator.com')
#         customer.add_email_address('fred.work@mailinator.com')
#         self.session.add(customer)
#         self.session.commit()
#         self.assertEqual(self.session.query(model.Customer).count(), 1)
#         self.assertEqual(self.session.query(model.CustomerEmailAddress).count(), 2)

#         while customer.emails:
#             customer.emails.pop()
#         self.session.commit()
#         self.assertEqual(self.session.query(model.Customer).count(), 1)
#         self.assertEqual(self.session.query(model.CustomerEmailAddress).count(), 0)

#         # changes weren't being recorded
#         self.assertEqual(self.session.query(model.Change).count(), 0)

#     def test_pop_with_changes(self):
#         record_changes(self.session)

#         customer = model.Customer()
#         customer.add_email_address('fred.home@mailinator.com')
#         customer.add_email_address('fred.work@mailinator.com')
#         self.session.add(customer)
#         self.session.commit()
#         self.assertEqual(self.session.query(model.Customer).count(), 1)
#         self.assertEqual(self.session.query(model.CustomerEmailAddress).count(), 2)

#         while customer.emails:
#             customer.emails.pop()
#         self.session.commit()
#         self.assertEqual(self.session.query(model.Customer).count(), 1)
#         self.assertEqual(self.session.query(model.CustomerEmailAddress).count(), 0)

#         # changes should have been recorded
#         changes = self.session.query(model.Change)
#         self.assertEqual(changes.count(), 3)

#         customer_change = changes.filter_by(class_name='Customer').one()
#         self.assertEqual(customer_change.uuid, customer.uuid)
#         self.assertFalse(customer_change.deleted)

#         email_changes = changes.filter_by(class_name='CustomerEmailAddress')
#         self.assertEqual(email_changes.count(), 2)
#         self.assertEqual([x.deleted for x in email_changes], [True, True])


class TestLabelProfile(DataTestCase):

    def test_get_printer_setting(self):
        profile = model.LabelProfile()
        self.session.add(profile)

        self.assertTrue(profile.uuid is None)
        setting = profile.get_printer_setting('some_setting')
        self.assertTrue(setting is None)
        self.assertTrue(profile.uuid is None)

        profile.uuid = 'some_uuid'
        self.session.add(model.Setting(
                name='labels.some_uuid.printer.some_setting',
                value='some_value'))
        self.session.flush()
        setting = profile.get_printer_setting('some_setting')
        self.assertEquals(setting, 'some_value')

    def test_save_printer_setting(self):
        self.assertEqual(self.session.query(model.Setting).count(), 0)
        profile = model.LabelProfile()
        self.session.add(profile)

        self.assertTrue(profile.uuid is None)
        profile.save_printer_setting('some_setting', 'some_value')
        self.assertFalse(profile.uuid is None)
        self.assertEqual(self.session.query(model.Setting).count(), 1)

        profile.uuid = 'some_uuid'
        profile.save_printer_setting('some_setting', 'some_value')
        self.assertEqual(self.session.query(model.Setting).count(), 2)
        setting = self.session.query(model.Setting)\
            .filter_by(name='labels.some_uuid.printer.some_setting')\
            .one()
        self.assertEqual(setting.value, 'some_value')



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import os
import warnings
from unittest import TestCase

import sqlalchemy as sa
from sqlalchemy.exc import SAWarning

from rattail.config import make_config
from rattail.db import model
from rattail.db import Session


warnings.filterwarnings(
    'ignore',
    r"^Dialect sqlite\+pysqlite does \*not\* support Decimal objects natively\, "
    "and SQLAlchemy must convert from floating point - rounding errors and other "
    "issues may occur\. Please consider storing Decimal numbers as strings or "
    "integers on this platform for lossless storage\.$",
    SAWarning, r'^sqlalchemy\..*$')


class NullProgress(object):
    """
    Dummy progress bar which does nothing, but allows for better test coverage
    when used with code under test.
    """

    def __init__(self, message, count):
        pass

    def update(self, value):
        pass

    def destroy(self):
        pass


class RattailMixin(object):
    """
    Generic mixin for ``TestCase`` classes which need common Rattail setup
    functionality.
    """
    engine_url = os.environ.get('RATTAIL_TEST_ENGINE_URL', 'sqlite://')
    host_engine_url = os.environ.get('RATTAIL_TEST_HOST_ENGINE_URL')

    def postgresql(self):
        return self.config.rattail_engine.url.get_dialect().name == 'postgresql'

    def setUp(self):
        self.setup_rattail()

    def tearDown(self):
        self.teardown_rattail()

    def setup_rattail(self):
        config = self.make_rattail_config()
        self.config = config
        self.rattail_config = config

        engine = sa.create_engine(self.engine_url)
        config.rattail_engines['default'] = engine
        config.rattail_engine = engine

        if self.host_engine_url:
            config.rattail_engines['host'] = sa.create_engine(self.host_engine_url)

        model = self.get_rattail_model()
        model.Base.metadata.create_all(bind=engine)

        Session.configure(bind=engine, rattail_config=config)
        self.session = Session()

    def teardown_rattail(self):
        self.session.close()
        Session.configure(bind=None, rattail_config=None)
        model = self.get_rattail_model()
        model.Base.metadata.drop_all(bind=self.rattail_config.rattail_engine)

    def make_rattail_config(self, **kwargs):
        kwargs.setdefault('files', [])
        return make_config(**kwargs)

    def get_rattail_model(self):
        return model


class RattailTestCase(RattailMixin, TestCase):
    """
    Generic base class for Rattail tests.
    """


class DataTestCase(TestCase):

    engine_url = os.environ.get('RATTAIL_TEST_ENGINE_URL', 'sqlite://')

    def setUp(self):
        self.engine = sa.create_engine(self.engine_url)
        model.Base.metadata.create_all(bind=self.engine)
        Session.configure(bind=self.engine)
        self.session = Session()
        self.extra_setup()

    def extra_setup(self):
        """
        Derivative classes may define this as necessary, to avoid having to
        override the :meth:`setUp()` method.
        """

    def tearDown(self):
        self.session.close()
        Session.configure(bind=None)
        model.Base.metadata.drop_all(bind=self.engine)

        # # TODO: This doesn't seem to be necessary, hopefully that's good?
        # for table in list(model.Base.metadata.sorted_tables):
        #     if table.name.startswith('batch.'):
        #         model.Base.metadata.remove(table)

        # TODO: Unfortunately this *does* seem to be necessary...
        model.Batch._rowclasses.clear()



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

from unittest import TestCase

from mock import Mock

from rattail.db import model
from rattail.db.model import people
from rattail.tests import DataTestCase


class TestPerson(DataTestCase):

    def test_unicode(self):
        person = model.Person()
        self.assertEqual(unicode(person), "")

        person = model.Person(display_name="Fred Flintstone")
        self.assertEqual(unicode(person), "Fred Flintstone")

    # TODO: this is duplicated in TestPerson
    def test_add_email_address(self):
        person = model.Person()
        self.assertEqual(len(person.emails), 0)
        person.add_email_address('fred@mailinator.com')
        self.assertEqual(len(person.emails), 1)
        email = person.emails[0]
        self.assertEqual(email.type, 'Home')

        person = model.Person()
        self.assertEqual(len(person.emails), 0)
        person.add_email_address('fred@mailinator.com', type='Work')
        self.assertEqual(len(person.emails), 1)
        email = person.emails[0]
        self.assertEqual(email.type, 'Work')

    # TODO: this is duplicated in TestPerson
    def test_add_phone_number(self):
        person = model.Person()
        self.assertEqual(len(person.phones), 0)
        person.add_phone_number('417-555-1234')
        self.assertEqual(len(person.phones), 1)
        phone = person.phones[0]
        self.assertEqual(phone.type, 'Home')

        person = model.Person()
        self.assertEqual(len(person.phones), 0)
        person.add_phone_number('417-555-1234', type='Work')
        self.assertEqual(len(person.phones), 1)
        phone = person.phones[0]
        self.assertEqual(phone.type, 'Work')


# TODO: deprecate/remove this?
class TestFunctions(TestCase):

    def test_get_person_display_name(self):
        name = people.get_person_display_name("Fred", "Flintstone")
        self.assertEqual(name, "Fred Flintstone")

    def test_get_person_display_name_from_context(self):
        context = Mock(current_parameters={'first_name': "Fred", 'last_name': "Flintstone"})
        name = people.get_person_display_name_from_context(context)
        self.assertEqual(name, "Fred Flintstone")



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Console Commands
"""

from __future__ import unicode_literals, absolute_import

import os
import sys
import time
import platform
import argparse
import datetime
import socket
import shutil
import warnings
import logging
from getpass import getpass

from rattail import __version__
from rattail.util import load_entry_points, load_object
from rattail.console import Progress
from rattail.config import make_config


log = logging.getLogger(__name__)


class ArgumentParser(argparse.ArgumentParser):
    """
    Custom argument parser.

    This overrides some of the parsing logic which is specific to the primary
    command object.
    """

    def parse_args(self, args=None, namespace=None):
        args, argv = self.parse_known_args(args, namespace)
        args.argv = argv
        return args


def date_argument(string):
    """
    Validate and coerce a date argument.

    This function is designed be used as the ``type`` parameter when calling
    ``ArgumentParser.add_argument()``, e.g.::

       parser = ArgumentParser()
       parser.add_argument('--date', type=date_argument)
    """
    try:
        date = datetime.datetime.strptime(string, '%Y-%m-%d').date()
    except ValueError:
        raise argparse.ArgumentTypeError("Date must be in YYYY-MM-DD format")
    return date


class Command(object):
    """
    The primary command for the application.

    This effectively *is* the ``rattail`` console application.  It mostly
    provides the structure for subcommands, which really do all the work.

    This command is designed to be subclassed, should your application need
    similar functionality.
    """
    name = 'rattail'
    version = __version__
    description = "Retail Software Framework"
    long_description = """
Rattail is a retail software framework.

Copyright (c) 2010-2015 Lance Edgar <lance@edbob.org>

This program comes with ABSOLUTELY NO WARRANTY.  This is free software,
and you are welcome to redistribute it under certain conditions.
See the file COPYING.txt for more information.
"""

    stdout = sys.stdout
    stderr = sys.stderr

    def __init__(self):
        self.subcommands = load_entry_points('{}.commands'.format(self.name.replace('-', '_')))

    def __unicode__(self):
        return unicode(self.name)

    @property
    def db_config_section(self):
        """
        Name of section in config file which should have database connection
        info.  This defaults to ``'rattail.db'`` but may be overridden so the
        command framework can sit in front of a non-Rattail database if needed.

        This is used to auto-configure a "default" database engine for the app,
        when any command is invoked.
        """
        return 'rattail.db'

    @property
    def db_session_factory(self):
        """
        Reference to the "primary" ``Session`` class, which will be configured
        automatically during app startup.  Defaults to :class:`rattail.db.Session`.
        """
        from rattail.db import Session
        return Session

    @property
    def db_model(self):
        """
        Reference to the Python module which is to be used as the primary data
        model.  Defaults to ``rattail.db.model``.
        """
        from rattail.db import model
        return model

    def iter_subcommands(self):
        """
        Iterate over the subcommands.

        This is a generator which yields each associated :class:`Subcommand`
        class sorted by :attr:`Subcommand.name`.
        """
        for name in sorted(self.subcommands):
            yield self.subcommands[name]

    def print_help(self):
        """
        Print help text for the primary command.

        The output will include a list of available subcommands.
        """
        self.stdout.write("""{0}

Usage: {1} [options] <command> [command-options]

Options:
  -c PATH, --config=PATH
                    Config path (may be specified more than once)
  -n, --no-init     Don't load config before executing command
  -P, --progress    Show progress indicators (where relevant)
  -V, --version     Display program version and exit

Commands:\n""".format(self.description, self.name))

        for cmd in self.iter_subcommands():
            self.stdout.write("  {0:<16s}  {1}\n".format(cmd.name, cmd.description))

        self.stdout.write("\nTry '{0} help <command>' for more help.\n".format(self.name))

    def run(self, *args):
        """
        Parse command line arguments and execute appropriate subcommand.
        """
        parser = ArgumentParser(
            prog=self.name,
            description=self.description,
            add_help=False,
            )

        parser.add_argument('-c', '--config', action='append', dest='config_paths',
                            metavar='PATH')
        parser.add_argument('-n', '--no-init', action='store_true', default=False)
        parser.add_argument('--no-extend-config', dest='extend_config', action='store_false')
        parser.add_argument('-P', '--progress', action='store_true', default=False)
        parser.add_argument('--stdout', metavar='PATH', type=argparse.FileType('w'),
                            help="Optional path to which STDOUT should be effectively redirected.")
        parser.add_argument('--stderr', metavar='PATH', type=argparse.FileType('w'),
                            help="Optional path to which STDERR should be effectively redirected.")
        parser.add_argument('-V', '--version', action='version',
                            version="%(prog)s {0}".format(self.version))
        parser.add_argument('command', nargs='*')

        # Parse args and determine subcommand.
        args = parser.parse_args(list(args))
        if not args or not args.command:
            self.print_help()
            return

        # Show (sub)command help if so instructed, or unknown subcommand.
        cmd = args.command.pop(0)
        if cmd == 'help':
            if len(args.command) != 1:
                self.print_help()
                return
            cmd = args.command[0]
            if cmd not in self.subcommands:
                self.print_help()
                return
            cmd = self.subcommands[cmd](parent=self)
            cmd.parser.print_help()
            return
        elif cmd not in self.subcommands:
            self.print_help()
            return

        # Okay, we should be done needing to print help messages.  Now it's
        # safe to redirect STDOUT/STDERR, if necessary.
        if args.stdout:
            self.stdout = args.stdout
        if args.stderr:
            self.stderr = args.stderr

        # Make the config object, and configure logging (or not).
        if args.no_init:
            config = make_config([], extend=False)
        else:
            logging.basicConfig()
            config = make_config(args.config_paths, extend=args.extend_config)

        # And finally, do something of real value...
        cmd = self.subcommands[cmd](self, config)
        cmd.show_progress = args.progress
        cmd.progress = Progress if args.progress else None
        cmd._run(*(args.command + args.argv))


class Subcommand(object):
    """
    Base class for application subcommands.
    """
    name = 'UNDEFINED'
    description = 'UNDEFINED'

    def __init__(self, parent=None, config=None, show_progress=None):
        self.parent = parent
        self.config = config
        self.stdout = getattr(parent, 'stdout', sys.stdout)
        self.stderr = getattr(parent, 'stderr', sys.stderr)
        self.show_progress = show_progress
        self.progress = Progress if show_progress else None
        self.parser = argparse.ArgumentParser(
            prog='{0} {1}'.format(getattr(self.parent, 'name', 'UNDEFINED'), self.name),
            description=self.description)
        self.add_parser_args(self.parser)

    def __repr__(self):
        return "Subcommand(name={0})".format(repr(self.name))

    def add_parser_args(self, parser):
        """
        Configure additional arguments for the subcommand argument parser.
        """
        pass
            
    def _run(self, *args):
        args = self.parser.parse_args(list(args))
        return self.run(args)

    def run(self, args):
        """
        Run the subcommand logic.
        """
        raise NotImplementedError


class AddUser(Subcommand):
    """
    Adds a user to the database.
    """

    name = 'adduser'
    description = "Add a user to the database."

    def add_parser_args(self, parser):
        parser.add_argument('username',
                            help="Username for the new account.")
        parser.add_argument('-A', '--administrator',
                            action='store_true',
                            help="Add the new user to the Administrator role.")

    def run(self, args):
        from rattail.db import Session
        from rattail.db import model
        from rattail.db.auth import set_user_password, administrator_role

        session = Session()

        if session.query(model.User).filter_by(username=args.username).count():
            session.close()
            self.stderr.write("User '{0}' already exists.\n".format(args.username))
            return

        passwd = ''
        while not passwd:
            try:
                passwd = getpass("Enter a password for user '{0}': ".format(args.username))
            except KeyboardInterrupt:
                self.stderr.write("\nOperation was canceled.\n")
                return

        user = model.User(username=args.username)
        set_user_password(user, passwd)
        if args.administrator:
            user.roles.append(administrator_role(session))
        session.add(user)
        session.commit()
        session.close()
        self.stdout.write("Created user: {0}\n".format(args.username))


class CloneDatabase(Subcommand):
    """
    Clone data from any source database into any target database.
    """
    name = 'clonedb'
    description = "Clone data from a source to a target database."

    def add_parser_args(self, parser):
        parser.add_argument('source_engine',
                            help="SQLAlchemy engine URL for the source database.")
        parser.add_argument('target_engine',
                            help="SQLAlchemy engine URL for the target database.")
        parser.add_argument('-m', '--model', default='rattail.db.model',
                            help="Dotted path of Python module which contains the data model.")
        parser.add_argument('-C', '--classes', nargs='*',
                            help="Model classes which should be cloned.  Possible values here "
                            "depends on which module contains the data model.  If no classes "
                            "are specified, all available will be cloned.")

    def run(self, args):
        from sqlalchemy import create_engine, orm
        from rattail.util import import_module_path

        model = import_module_path(args.model)
        classes = args.classes
        assert classes

        source_engine = create_engine(args.source_engine)
        target_engine = create_engine(args.target_engine)
        model.Base.metadata.drop_all(bind=target_engine)
        model.Base.metadata.create_all(bind=target_engine)

        Session = orm.sessionmaker()
        src_session = Session(bind=source_engine)
        dst_session = Session(bind=target_engine)

        for clsname in classes:
            log.info("cloning data for model: {0}".format(clsname))
            cls = getattr(model, clsname)
            src_query = src_session.query(cls)
            count = src_query.count()
            log.debug("found {0} {1} records to clone".format(count, clsname))
            if not count:
                continue

            mapper = orm.class_mapper(cls)
            key_query = src_session.query(*mapper.primary_key)

            prog = None
            if self.progress:
                prog = self.progress("Cloning data for model: {0}".format(clsname), count)
            for i, key in enumerate(key_query, 1):

                src_instance = src_query.get(key)
                dst_session.merge(src_instance)
                dst_session.flush()

                if prog:
                    prog.update(i)
            if prog:
                prog.destroy()

        src_session.close()
        dst_session.commit()
        dst_session.close()


class DataSync(Subcommand):
    """
    Interacts with the datasync daemon.  This command expects a subcommand; one
    of the following:

    * ``rattail datasync start``
    * ``rattail datasync stop``
    """
    name = 'datasync'
    description = "Manage the DataSync daemon"

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        start = subparsers.add_parser('start', help="Start service")
        start.set_defaults(subcommand='start')
        stop = subparsers.add_parser('stop', help="Stop service")
        stop.set_defaults(subcommand='stop')
        wait = subparsers.add_parser('wait', help="Wait for changes to be processed")
        wait.set_defaults(subcommand='wait')

        parser.add_argument('-p', '--pidfile', metavar='PATH', default='/var/run/rattail/datasync.pid',
                            help="Path to PID file.")
        parser.add_argument('-D', '--do-not-daemonize',
                            action='store_false', dest='daemonize', default=True,
                            help="Do not daemonize when starting.")
        parser.add_argument('-T', '--timeout', metavar='MINUTES', type=int, default=0,
                            help="Optional timeout (in minutes) for use with the 'wait' command.  "
                            "If specified, the waiting still stop after the given number of minutes "
                            "and exit with a nonzero code to indicate failure.")

    def run(self, args):
        from rattail.datasync.daemon import DataSyncDaemon

        if args.subcommand == 'wait':
            self.wait(args)

        else: # manage the daemon
            daemon = DataSyncDaemon(args.pidfile, config=self.config)
            if args.subcommand == 'stop':
                daemon.stop()
            else: # start
                try:
                    daemon.start(daemonize=args.daemonize)
                except KeyboardInterrupt:
                    if not args.daemonize:
                        self.stderr.write("Interrupted.\n")
                    else:
                        raise

    def wait(self, args):
        model = self.parent.db_model
        Session = self.parent.db_session_factory
        session = Session()
        started = datetime.datetime.utcnow()
        log.debug("will wait for current change queue to clear")
        last_logged = started

        changes = session.query(model.DataSyncChange)
        count = changes.count()
        log.debug("there are {} changes in the queue".format(count))
        while count:
            try:
                now = datetime.datetime.utcnow()

                if args.timeout and (now - started).seconds >= (args.timeout * 60):
                    log.warning("datasync wait timed out after {} minutes, with {} changes in queue".format(args.timeout, count))
                    sys.exit(1)

                if (now - last_logged).seconds >= 60:
                    log.debug("still waiting, {} changes in the datasync queue".format(count))
                    last_logged = now

                time.sleep(1)
                count = changes.count()

            except KeyboardInterrupt:
                self.stderr.write("Waiting cancelled by user\n")
                session.close()
                sys.exit(1)

        session.close()
        log.debug("all datasync changes have been processed")


class EmailBouncer(Subcommand):
    """
    Interacts with the email bouncer daemon.  This command expects a
    subcommand; one of the following:

    * ``rattail bouncer start``
    * ``rattail bouncer stop``
    """
    name = 'bouncer'
    description = "Manage the email bouncer daemon"

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        start = subparsers.add_parser('start', help="Start service")
        start.set_defaults(subcommand='start')
        stop = subparsers.add_parser('stop', help="Stop service")
        stop.set_defaults(subcommand='stop')

        parser.add_argument('-p', '--pidfile', metavar='PATH', default='/var/run/rattail/bouncer.pid',
                            help="Path to PID file.")
        parser.add_argument('-D', '--do-not-daemonize',
                            action='store_false', dest='daemonize', default=True,
                            help="Do not daemonize when starting.")

    def run(self, args):
        from rattail.bouncer.daemon import BouncerDaemon

        daemon = BouncerDaemon(args.pidfile, config=self.config)
        if args.subcommand == 'stop':
            daemon.stop()
        else: # start
            try:
                daemon.start(daemonize=args.daemonize)
            except KeyboardInterrupt:
                if not args.daemonize:
                    self.stderr.write("Interrupted.\n")
                else:
                    raise


class DateOrganize(Subcommand):
    """
    Organize files in a given directory, according to date.
    """
    name = 'date-organize'
    description = "Organize files in a given directory according to date."

    def add_parser_args(self, parser):
        parser.add_argument('folder', metavar='PATH',
                            help="Path to directory containing files which are "
                            "to be organized by date.")

    def run(self, args):
        today = datetime.date.today()
        for filename in sorted(os.listdir(args.folder)):
            path = os.path.join(args.folder, filename)
            if os.path.isfile(path):
                mtime = datetime.datetime.fromtimestamp(os.path.getmtime(path))
                if mtime.date() < today:
                    datedir = mtime.strftime(os.sep.join(('%Y', '%m', '%d')))
                    datedir = os.path.join(args.folder, datedir)
                    if not os.path.exists(datedir):
                        os.makedirs(datedir)
                    shutil.move(path, datedir)


class DatabaseSyncCommand(Subcommand):
    """
    Controls the database synchronization service.
    """

    name = 'dbsync'
    description = "Manage the database synchronization service"

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        start = subparsers.add_parser('start', help="Start service")
        start.set_defaults(subcommand='start')
        stop = subparsers.add_parser('stop', help="Stop service")
        stop.set_defaults(subcommand='stop')

        if sys.platform == 'linux2':
            parser.add_argument('-p', '--pidfile',
                                help="Path to PID file", metavar='PATH')
            parser.add_argument('-D', '--do-not-daemonize',
                                action='store_false', dest='daemonize', default=True,
                                help="Do not daemonize when starting.")

    def run(self, args):
        from rattail.db.sync import linux as dbsync

        if args.subcommand == 'start':
            try:
                dbsync.start_daemon(self.config, args.pidfile, args.daemonize)
            except KeyboardInterrupt:
                if not args.daemonize:
                    self.stderr.write("Interrupted.\n")
                else:
                    raise

        elif args.subcommand == 'stop':
            dbsync.stop_daemon(self.config, args.pidfile)


class Dump(Subcommand):
    """
    Do a simple data dump.
    """

    name = 'dump'
    description = "Dump data to file."

    def add_parser_args(self, parser):
        parser.add_argument(
            '--output', '-o', metavar='FILE',
            help="Optional path to output file.  If none is specified, "
            "data will be written to standard output.")
        parser.add_argument(
            'model', help="Model whose data will be dumped.")

    def get_model(self):
        """
        Returns the module which contains all relevant data models.

        By default this returns ``rattail.db.model``, but this method may be
        overridden in derived commands to add support for extra data models.
        """
        from rattail.db import model
        return model

    def run(self, args):
        from rattail.db import Session
        from rattail.db.dump import dump_data

        model = self.get_model()
        if hasattr(model, args.model):
            cls = getattr(model, args.model)
        else:
            self.stderr.write("Unknown model: {0}\n".format(args.model))
            sys.exit(1)

        progress = None
        if self.show_progress: # pragma no cover
            progress = Progress

        if args.output:
            output = open(args.output, 'wb')
        else:
            output = self.stdout

        session = Session()
        dump_data(session, cls, output, progress=progress)
        session.close()

        if output is not self.stdout:
            output.close()


class FileMonitorCommand(Subcommand):
    """
    Interacts with the file monitor service; called as ``rattail filemon``.
    This command expects a subcommand; one of the following:

    * ``rattail filemon start``
    * ``rattail filemon stop``

    On Windows platforms, the following additional subcommands are available:

    * ``rattail filemon install``
    * ``rattail filemon uninstall`` (or ``rattail filemon remove``)

    .. note::
       The Windows Vista family of operating systems requires you to launch
       ``cmd.exe`` as an Administrator in order to have sufficient rights to
       run the above commands.

    .. See :doc:`howto.use_filemon` for more information.
    """

    name = 'filemon'
    description = "Manage the file monitor service"

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        start = subparsers.add_parser('start', help="Start service")
        start.set_defaults(subcommand='start')
        stop = subparsers.add_parser('stop', help="Stop service")
        stop.set_defaults(subcommand='stop')

        if sys.platform == 'linux2':
            parser.add_argument('-p', '--pidfile',
                                help="Path to PID file.", metavar='PATH')
            parser.add_argument('-D', '--do-not-daemonize',
                                action='store_false', dest='daemonize', default=True,
                                help="Do not daemonize when starting.")

        elif sys.platform == 'win32': # pragma no cover

            install = subparsers.add_parser('install', help="Install service")
            install.set_defaults(subcommand='install')
            install.add_argument('-a', '--auto-start', action='store_true',
                                 help="Configure service to start automatically.")
            install.add_argument('-U', '--username',
                                 help="User account under which the service should run.")

            remove = subparsers.add_parser('remove', help="Uninstall (remove) service")
            remove.set_defaults(subcommand='remove')

            uninstall = subparsers.add_parser('uninstall', help="Uninstall (remove) service")
            uninstall.set_defaults(subcommand='remove')

    def run(self, args):
        if sys.platform == 'linux2':
            from rattail.filemon import linux as filemon

            if args.subcommand == 'start':
                filemon.start_daemon(self.config, args.pidfile, args.daemonize)

            elif args.subcommand == 'stop':
                filemon.stop_daemon(self.config, args.pidfile)

        elif sys.platform == 'win32': # pragma no cover
            self.run_win32(args)

        else:
            self.stderr.write("File monitor is not supported on platform: {0}\n".format(sys.platform))
            sys.exit(1)

    def run_win32(self, args): # pragma no cover
        from rattail.win32 import require_elevation
        from rattail.win32 import service
        from rattail.win32 import users
        from rattail.filemon import win32 as filemon

        require_elevation()

        options = []
        if args.subcommand == 'install':

            username = args.username
            if username:
                if '\\' in username:
                    server, username = username.split('\\')
                else:
                    server = socket.gethostname()
                if not users.user_exists(username, server):
                    sys.stderr.write("User does not exist: {0}\\{1}\n".format(server, username))
                    sys.exit(1)

                password = ''
                while password == '':
                    password = getpass("Password for service user: ").strip()
                options.extend(['--username', r'{0}\{1}'.format(server, username)])
                options.extend(['--password', password])

            if args.auto_start:
                options.extend(['--startup', 'auto'])

        service.execute_service_command(filemon, args.subcommand, *options)

        # If installing with custom user, grant "logon as service" right.
        if args.subcommand == 'install' and args.username:
            users.allow_logon_as_service(username)

        # TODO: Figure out if the following is even required, or if instead we
        # should just be passing '--startup delayed' to begin with?

        # If installing auto-start service on Windows 7, we should update
        # its startup type to be "Automatic (Delayed Start)".
        # TODO: Improve this check to include Vista?
        if args.subcommand == 'install' and args.auto_start:
            if platform.release() == '7':
                name = filemon.RattailFileMonitor._svc_name_
                service.delayed_auto_start_service(name)


class OldImportSubcommand(Subcommand):
    """
    Base class for subcommands which use the data importing system.
    """
    supports_versioning = True

    def add_parser_args(self, parser):
        handler = self.get_handler(quiet=True)
        if self.supports_versioning:
            parser.add_argument('--no-versioning', action='store_true',
                                help="Disables versioning during the import.  This is "
                                "intended to be useful e.g. during initial import, where "
                                "the process can be quite slow even without the overhead "
                                "of versioning.")
        parser.add_argument('--warnings', '-W', action='store_true',
                            help="Whether to log warnings if any data model "
                            "writes occur.  Intended to help stay in sync "
                            "with an external data source.")
        parser.add_argument('--max-updates', type=int,
                            help="Maximum number of record updates (or additions) which, if "
                            "reached, should cause the importer to stop early.  Note that the "
                            "updates which have completed will be committed unless a dry run "
                            "is in effect.")
        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the motions and allow logging to occur, "
                            "but do not actually commit the transaction at the end.")
        parser.add_argument('models', nargs='*', metavar='MODEL',
                            help="Which models to import.  If none are specified, all models will "
                            "be imported.  Or, specify only those you wish to import.  Supported "
                            "models are: {0}".format(', '.join(handler.get_importer_keys())))

    def run(self, args):
        log.info("begin {0} for data model(s): {1}".format(
                self.name, ', '.join(args.models or ["ALL"])))

        Session = self.parent.db_session_factory
        if self.supports_versioning:
            if args.no_versioning:
                from rattail.db.continuum  import disable_versioning
                disable_versioning()
            session = Session(continuum_user=self.continuum_user)
        else:
            session = Session()

        self.import_data(args, session)

        if args.dry_run:
            session.rollback()
            log.info("dry run, so transaction was rolled back")
        else:
            session.commit()
            log.info("transaction was committed")
        session.close()

    def get_handler_factory(self, quiet=False):
        """
        This method must return a factory, which will in turn generate a
        handler instance to be used by the command.  Note that you *must*
        override this method.
        """
        raise NotImplementedError

    def get_handler(self, **kwargs):
        """
        Returns a handler instance to be used by the command.
        """
        factory = self.get_handler_factory(quiet=kwargs.pop('quiet', False))
        return factory(getattr(self, 'config', None), **kwargs)

    @property
    def continuum_user(self):
        """
        Info needed to assign the Continuum user for the database session.
        """

    def import_data(self, args, session):
        """
        Perform a data import, with the given arguments and database session.
        """
        handler = self.get_handler(session=session)
        models = args.models or handler.get_importer_keys()
        updates = handler.import_data(models, max_updates=args.max_updates,
                                      progress=self.progress)
        if args.warnings and updates:
            handler.process_warnings(updates, command=self, models=models, dry_run=args.dry_run,
                                     render_record=self.get_record_renderer(),
                                     progress=self.progress)

    def get_record_renderer(self):
        """
        Get the record renderer for email notifications.  Note that config may
        override the default.
        """
        spec = self.config.get('{0}.{1}'.format(self.parent.name, self.name), 'record_renderer',
                               default='rattail.db.importing:RecordRenderer')
        return load_object(spec)(self.config)


class NewImportSubcommand(Subcommand):
    """
    Base class for subcommands which use the (new) data importing system.
    """

    def get_handler_factory(self, args=None):
        """
        This method must return a factory, which will in turn generate a
        handler instance to be used by the command.  Note that you *must*
        override this method.
        """
        raise NotImplementedError

    def get_handler(self, args=None, **kwargs):
        """
        Returns a handler instance to be used by the command.
        """
        factory = self.get_handler_factory(args)
        kwargs = self.get_handler_kwargs(args, **kwargs)
        kwargs['command'] = self
        return factory(getattr(self, 'config', None), **kwargs)

    def get_handler_kwargs(self, args, **kwargs):
        """
        Return a dict of kwargs to be passed to the handler factory.
        """
        return kwargs

    def add_parser_args(self, parser):
        handler = self.get_handler()

        # model names (aka importer keys)
        parser.add_argument('models', nargs='*', metavar='MODEL',
                            help="Which data models to import.  If you specify any, then only data "
                            "for those models will be imported.  If you do not specify any, then all "
                            "*default* models will be imported.  Supported models are: ({})".format(
                                ', '.join(handler.get_importer_keys())))

        # start/end date
        parser.add_argument('--start-date', type=date_argument,
                            help="Optional (inclusive) starting point for date range, by which host "
                            "data should be filtered.  Only used by certain importers.")
        parser.add_argument('--end-date', type=date_argument,
                            help="Optional (inclusive) ending point for date range, by which host "
                            "data should be filtered.  Only used by certain importers.")

        # allow create?
        parser.add_argument('--create', action='store_true', default=True,
                            help="Allow new records to be created during the import.")
        parser.add_argument('--no-create', action='store_false', dest='create',
                            help="Do not allow new records to be created during the import.")
        parser.add_argument('--max-create', type=int, metavar='COUNT',
                            help="Maximum number of records which may be created, after which a "
                            "given import task should stop.  Note that this applies on a per-model "
                            "basis and not overall.")

        # allow update?
        parser.add_argument('--update', action='store_true', default=True,
                            help="Allow existing records to be updated during the import.")
        parser.add_argument('--no-update', action='store_false', dest='update',
                            help="Do not allow existing records to be updated during the import.")
        parser.add_argument('--max-update', type=int, metavar='COUNT',
                            help="Maximum number of records which may be updated, after which a "
                            "given import task should stop.  Note that this applies on a per-model "
                            "basis and not overall.")

        # allow delete?
        parser.add_argument('--delete', action='store_true', default=False,
                            help="Allow records to be deleted during the import.")
        parser.add_argument('--no-delete', action='store_false', dest='delete',
                            help="Do not allow records to be deleted during the import.")
        parser.add_argument('--max-delete', type=int, metavar='COUNT',
                            help="Maximum number of records which may be deleted, after which a "
                            "given import task should stop.  Note that this applies on a per-model "
                            "basis and not overall.")

        # max total changes, per model
        parser.add_argument('--max-total', type=int, metavar='COUNT',
                            help="Maximum number of *any* record changes which may occur, after which "
                            "a given import task should stop.  Note that this applies on a per-model "
                            "basis and not overall.")

        # treat changes as warnings?
        parser.add_argument('--warnings', '-W', action='store_true',
                            help="Set this flag if you expect a \"clean\" import, and wish for any "
                            "changes which do occur to be processed further and/or specially.  The "
                            "behavior of this flag is ultimately up to the import handler, but the "
                            "default is to send an email notification.")

        # dry run?
        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the full motions and allow logging etc. to "
                            "occur, but rollback (abort) the transaction at the end.")

    def run(self, args):
        log.info("begin `{} {}` for data models: {}".format(
                self.parent.name, self.name, ', '.join(args.models or ["(ALL)"])))

        Session = self.parent.db_session_factory
        session = Session()

        self.import_data(args, session)

        if args.dry_run:
            session.rollback()
            log.info("dry run, so transaction was rolled back")
        else:
            session.commit()
            log.info("transaction was committed")
        session.close()

    def import_data(self, args, session):
        """
        Perform a data import, with the given arguments and database session.
        """
        handler = self.get_handler(args=args, session=session, progress=self.progress)
        models = args.models or handler.get_default_keys()
        log.debug("using handler: {}".format(handler))
        log.debug("importing models: {}".format(models))
        log.debug("args are: {}".format(args))
        handler.import_data(models, args)


class ImportCSV(OldImportSubcommand):
    """
    Import data from a CSV file.
    """
    name = 'import-csv'
    description = "Import data from a CSV file."

    def add_parser_args(self, parser):
        super(ImportCSV, self).add_parser_args(parser)
        parser.add_argument('importer',
                            help="Spec string for importer class which should handle the import.")
        parser.add_argument('csv_path',
                            help="Path to the data file which will be imported.")

    def import_data(self, args, session):
        from rattail.db.importing.providers.csv import make_provider

        provider = make_provider(self.config, session, args.importer, data_path=args.csv_path)
        data = provider.get_data(progress=self.progress)
        affected = provider.importer.import_data(data, provider.supported_fields, 'uuid',
                                                 progress=self.progress)
        log.info("added or updated {0} {1} records".format(affected, provider.model_name))


class InitializeDatabase(Subcommand):
    """
    Creates the initial Rattail tables within a database.
    """

    name = 'initdb'
    description = "Create initial tables in a database."

    def add_parser_args(self, parser):
        parser.add_argument('--with-admin', action='store_true',
                            help="Create an 'admin' user account if none exists.")

    def run(self, args):
        from alembic.util import obfuscate_url_pw
        from rattail.db import model, auth, Session

        # Maybe create 'admin' user with 'admin' password.
        if args.with_admin:
            session = Session()
            if not session.query(model.User).filter_by(username='admin').count():
                admin = model.User(username='admin')
                auth.set_user_password(admin, 'admin')
                admin.roles.append(auth.administrator_role(session))
                session.add(admin)
                self.stdout.write("[rattail] created 'admin' user with password 'admin'\n")
                session.commit()
            session.close()

        self.stdout.write("[rattail] initialized schema for database: {0}\n".format(
            obfuscate_url_pw(self.config.rattail_engine.url)))


class LoadHostDataCommand(Subcommand):
    """
    Loads data from the Rattail host database, if one is configured.
    """

    name = 'load-host-data'
    description = "Load data from host database"

    def run(self, args):
        from .db import get_engines
        from .db import load

        engines = get_engines(self.config)
        if 'host' not in engines:
            sys.stderr.write("Host engine URL not configured.\n")
            sys.exit(1)

        proc = load.LoadProcessor(self.config)
        proc.load_all_data(engines['host'], Progress)


class MakeUserCommand(Subcommand):
    """
    Creates a system user for Rattail.
    """
    name = 'make-user'
    description = "Create a system user account for Rattail"

    def add_parser_args(self, parser):
        parser.add_argument('-U', '--username', metavar='USERNAME', default='rattail',
                            help="Username for the new user; defaults to 'rattail'.")
        parser.add_argument('--full-name', metavar='FULL_NAME',
                            help="Full (display) name for the new user.")
        parser.add_argument('--comment', metavar='COMMENT',
                            help="Comment string for the new user.")

    def run(self, args):
        if sys.platform != 'win32':
            sys.stderr.write("Sorry, only win32 platform is supported.\n")
            sys.exit(1)

        from rattail.win32 import users
        from rattail.win32 import require_elevation

        require_elevation()

        if users.user_exists(args.username):
            sys.stderr.write("User already exists: {0}\n".format(args.username))
            sys.exit(1)

        try:
            password = None
            while not password:
                password = getpass(b"Enter a password: ").strip()
        except KeyboardInterrupt:
            sys.stderr.write("Operation canceled by user.")
            sys.exit(2)

        users.create_user(args.username, password,
                          full_name=args.full_name, comment=args.comment)
        sys.stdout.write("Created user: {0}\n".format(args.username))


class PalmCommand(Subcommand):
    """
    Manages registration for the HotSync Manager conduit; called as::

       rattail palm
    """

    name = 'palm'
    description = "Manage the HotSync Manager conduit registration"

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        register = subparsers.add_parser('register', help="Register Rattail conduit")
        register.set_defaults(subcommand='register')

        unregister = subparsers.add_parser('unregister', help="Unregister Rattail conduit")
        unregister.set_defaults(subcommand='unregister')

    def run(self, args):
        from rattail import palm
        from rattail.win32 import require_elevation
        from rattail.exceptions import PalmError

        require_elevation()

        if args.subcommand == 'register':
            try:
                palm.register_conduit()
            except PalmError, error:
                sys.stderr.write(str(error) + '\n')

        elif args.subcommand == 'unregister':
            try:
                palm.unregister_conduit()
            except PalmError, error:
                sys.stderr.write(str(error) + '\n')
                

class PurgeBatchesCommand(Subcommand):
    """
    Purges stale batches from the database; called as:

    .. code-block:: sh

       rattail purge-batches
    """
    name = 'purge-batches'
    description = "Purge stale batches from the database"

    def add_parser_args(self, parser):
        parser.add_argument('-A', '--all', action='store_true',
                            help="Purge ALL batches regardless of purge date")
        parser.add_argument('--date', '-D', type=date_argument,
                            help="Optional effective date for the purge.  If "
                            "none is specified, the current date is assumed.")

    def run(self, args):
        from alembic.util import obfuscate_url_pw
        from rattail.db import Session
        from rattail.db.batches import util

        log.info("purging batches from database: {0}".format(obfuscate_url_pw(Session.kw['bind'].url)))
        normal = util.purge_batches(effective_date=args.date, purge_everything=args.all)
        orphaned = util.purge_orphaned_batches()
        log.info("purged {0} normal and {1} orphaned batches".format(normal, orphaned))


def main(*args):
    """
    The primary entry point for the Rattail command system.
    """
    if args:
        args = list(args)
    else:
        args = sys.argv[1:]

    cmd = Command()
    cmd.run(*args)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

from unittest import TestCase

from sqlalchemy import orm
from mock import patch, DEFAULT, Mock, MagicMock, call

from rattail import db
from rattail.db import changes, model
from rattail.tests import DataTestCase
from rattail.config import RattailConfig
from rattail.core import get_uuid


class TestRecordChangesFunc(TestCase):

    def test_session_class(self):
        Session = orm.sessionmaker()
        self.assertRaises(KeyError, Session.kw.__getitem__, 'rattail_record_changes')
        self.assertRaises(AttributeError, getattr, Session, 'rattail_record_changes')
        changes.record_changes(Session)
        self.assertTrue(Session.rattail_record_changes)

    def test_session_instance(self):
        session = db.Session()
        self.assertFalse(session.rattail_record_changes)
        changes.record_changes(session)
        self.assertTrue(session.rattail_record_changes)
        session.close()

    def test_recorder(self):

        # no recorder
        session = db.Session()
        self.assertRaises(AttributeError, getattr, session, 'rattail_change_recorder')
        session.close()

        # default recorder
        session = db.Session()
        changes.record_changes(session)
        self.assertIs(type(session.rattail_change_recorder), changes.ChangeRecorder)
        session.close()

        # specify recorder instance
        recorder = changes.ChangeRecorder()
        session = db.Session()
        changes.record_changes(session, recorder=recorder)
        self.assertIs(session.rattail_change_recorder, recorder)
        session.close()

        # specify recorder factory
        session = db.Session()
        changes.record_changes(session, recorder=changes.ChangeRecorder)
        self.assertIs(type(session.rattail_change_recorder), changes.ChangeRecorder)
        session.close()

        # specify recorder spec via config
        config = RattailConfig()
        config.set('rattail.db', 'changes.recorder', 'rattail.db.changes:ChangeRecorder')
        session = db.Session()
        changes.record_changes(session, config=config)
        self.assertIs(type(session.rattail_change_recorder), changes.ChangeRecorder)
        session.close()

        # invalid recorder
        session = db.Session()
        self.assertRaises(ValueError, changes.record_changes, session, recorder='bogus')
        session.close()


class TestChangeRecorder(DataTestCase):

    def test_ignore_object(self):
        recorder = changes.ChangeRecorder()
        self.assertTrue(recorder.ignore_object(model.Setting()))
        self.assertTrue(recorder.ignore_object(model.Change()))
        self.assertTrue(recorder.ignore_object(model.DataSyncChange()))
        self.assertFalse(recorder.ignore_object(model.Product()))
        self.assertFalse(recorder.ignore_object(model.Customer()))

    def test_process_new_object(self):
        self.assertEqual(self.session.query(model.Change).count(), 0)
        recorder = changes.ChangeRecorder()
        product = model.Product(uuid='6de299ca178d11e6be2c3ca9f40bc550')
        recorder.process_new_object(self.session, product)
        change = self.session.query(model.Change).one()
        self.assertEqual(change.class_name, 'Product')
        self.assertEqual(change.object_key, '6de299ca178d11e6be2c3ca9f40bc550')
        self.assertFalse(change.deleted)

    def test_process_dirty_object(self):
        self.assertEqual(self.session.query(model.Change).count(), 0)
        recorder = changes.ChangeRecorder()
        product = model.Product(uuid='6de299ca178d11e6be2c3ca9f40bc550')
        recorder.process_dirty_object(self.session, product)
        change = self.session.query(model.Change).one()
        self.assertEqual(change.class_name, 'Product')
        self.assertEqual(change.object_key, '6de299ca178d11e6be2c3ca9f40bc550')
        self.assertFalse(change.deleted)

    def test_process_deleted_object(self):
        self.assertEqual(self.session.query(model.Change).count(), 0)
        recorder = changes.ChangeRecorder()
        product = model.Product(uuid='6de299ca178d11e6be2c3ca9f40bc550')
        recorder.process_deleted_object(self.session, product)
        change = self.session.query(model.Change).one()
        self.assertEqual(change.class_name, 'Product')
        self.assertEqual(change.object_key, '6de299ca178d11e6be2c3ca9f40bc550')
        self.assertTrue(change.deleted)

    def test_process_deleted_object_special(self):
        self.assertEqual(self.session.query(model.Change).count(), 0)
        recorder = changes.ChangeRecorder()

        person = model.Person(uuid='06100a34178e11e6a8633ca9f40bc550')
        for Model in (model.PersonEmailAddress, model.PersonPhoneNumber, model.PersonMailingAddress):

            self.assertEqual(self.session.query(model.Change).count(), 0)
            uuid = get_uuid()
            obj = Model(uuid=uuid, person=person)
            recorder.process_deleted_object(self.session, obj)
            self.assertEqual(self.session.query(model.Change).count(), 2)

            change = self.session.query(model.Change).filter_by(class_name=Model.__name__).one()
            self.assertEqual(change.object_key, uuid)
            self.assertTrue(change.deleted)

            change = self.session.query(model.Change).filter_by(class_name='Person').one()
            self.assertEqual(change.object_key, '06100a34178e11e6a8633ca9f40bc550')
            self.assertFalse(change.deleted)

            self.session.query(model.Change).delete(synchronize_session=False)

        employee = model.Employee(uuid='b302ac38178e11e6b8843ca9f40bc550')
        for Model in (model.EmployeeStore, model.EmployeeDepartment):

            self.assertEqual(self.session.query(model.Change).count(), 0)
            uuid = get_uuid()
            obj = Model(uuid=uuid, employee=employee)
            recorder.process_deleted_object(self.session, obj)
            self.assertEqual(self.session.query(model.Change).count(), 2)

            change = self.session.query(model.Change).filter_by(class_name=Model.__name__).one()
            self.assertEqual(change.object_key, uuid)
            self.assertTrue(change.deleted)

            change = self.session.query(model.Change).filter_by(class_name='Employee').one()
            self.assertEqual(change.object_key, 'b302ac38178e11e6b8843ca9f40bc550')
            self.assertFalse(change.deleted)

            self.session.query(model.Change).delete(synchronize_session=False)

    def test_record_change(self):
        self.assertEqual(self.session.query(model.Change).count(), 0)
        recorder = changes.ChangeRecorder()

        recorder.record_change(self.session, class_name='Bogus', object_key='bogus', deleted=False)
        self.assertEqual(self.session.query(model.Change).count(), 1)
        change = self.session.query(model.Change).one()
        self.assertEqual(change.class_name, 'Bogus')
        self.assertEqual(change.object_key, 'bogus')
        self.assertFalse(change.deleted)

        recorder.record_change(self.session, class_name='Invalid', object_key='invalid', deleted=True)
        self.assertEqual(self.session.query(model.Change).count(), 2)
        change = self.session.query(model.Change).filter_by(class_name='Invalid').one()
        self.assertEqual(change.object_key, 'invalid')
        self.assertTrue(change.deleted)

    def test_record_rattail_change(self):
        self.assertEqual(self.session.query(model.Change).count(), 0)
        recorder = changes.ChangeRecorder()

        # ignore change object (TODO: redundant?)
        self.assertFalse(recorder.record_rattail_change(self.session, model.Change()))
        self.assertFalse(recorder.record_rattail_change(self.session, model.DataSyncChange()))
        self.assertEqual(self.session.query(model.Change).count(), 0)

        # ignore batch object (TODO: redundant?)
        self.assertFalse(recorder.record_rattail_change(self.session, model.Batch()))
        self.assertFalse(recorder.record_rattail_change(self.session, model.BatchColumn()))
        self.assertFalse(recorder.record_rattail_change(self.session, model.BatchRow()))
        self.assertFalse(recorder.record_rattail_change(self.session, model.BatchMixin()))
        self.assertFalse(recorder.record_rattail_change(self.session, model.BatchRowMixin()))
        self.assertEqual(self.session.query(model.Change).count(), 0)

        # ignore instance with no UUID attribute
        self.assertFalse(recorder.record_rattail_change(self.session, model.Setting()))
        self.assertFalse(recorder.record_rattail_change(self.session, model.Permission()))
        self.assertEqual(self.session.query(model.Change).count(), 0)

        product = model.Product(uuid='c1b0fb94178f11e680fd3ca9f40bc550')

        # default
        self.assertTrue(recorder.record_rattail_change(self.session, product))
        change = self.session.query(model.Change).one()
        self.assertEqual(change.class_name, 'Product')
        self.assertEqual(change.object_key, 'c1b0fb94178f11e680fd3ca9f40bc550')
        self.assertFalse(change.deleted)
        self.session.query(model.Change).delete(synchronize_session=False)

        # new
        self.assertTrue(recorder.record_rattail_change(self.session, product, type_='new'))
        change = self.session.query(model.Change).one()
        self.assertEqual(change.class_name, 'Product')
        self.assertEqual(change.object_key, 'c1b0fb94178f11e680fd3ca9f40bc550')
        self.assertFalse(change.deleted)
        self.session.query(model.Change).delete(synchronize_session=False)

        # dirty
        self.assertTrue(recorder.record_rattail_change(self.session, product, type_='dirty'))
        change = self.session.query(model.Change).one()
        self.assertEqual(change.class_name, 'Product')
        self.assertEqual(change.object_key, 'c1b0fb94178f11e680fd3ca9f40bc550')
        self.assertFalse(change.deleted)
        self.session.query(model.Change).delete(synchronize_session=False)

        # deleted
        self.assertTrue(recorder.record_rattail_change(self.session, product, type_='deleted'))
        change = self.session.query(model.Change).one()
        self.assertEqual(change.class_name, 'Product')
        self.assertEqual(change.object_key, 'c1b0fb94178f11e680fd3ca9f40bc550')
        self.assertTrue(change.deleted)
        self.session.query(model.Change).delete(synchronize_session=False)


class TestChangeRecorderLegacy(TestCase):

    def test_init(self):
        recorder = changes.ChangeRecorder()

    # def test_record_change(self):
    #     session = Mock()
    #     recorder = changes.ChangeRecorder()
    #     recorder.ensure_uuid = Mock()

    #     # don't record changes for changes
    #     self.assertFalse(recorder.record_change(session, model.Change()))

    #     # don't record changes for batch data
    #     self.assertFalse(recorder.record_change(session, model.Batch()))
    #     self.assertFalse(recorder.record_change(session, model.BatchColumn()))
    #     self.assertFalse(recorder.record_change(session, model.BatchRow()))

    #     # don't record changes for objects with no uuid attribute
    #     self.assertFalse(recorder.record_change(session, object()))

    #     # none of the above should have involved a call to `ensure_uuid()`
    #     self.assertFalse(recorder.ensure_uuid.called)

    #     # so far no *new* changes have been created
    #     self.assertFalse(session.add.called)

    #     # mock up session to force new change creation
    #     session.query.return_value = session
    #     session.get.return_value = None
    #     self.assertTrue(recorder.record_change(session, model.Product()))

    @patch.multiple('rattail.db.changes', get_uuid=DEFAULT, object_mapper=DEFAULT)
    def test_ensure_uuid(self, get_uuid, object_mapper):
        recorder = changes.ChangeRecorder()
        uuid_column = Mock()
        object_mapper.return_value.columns.__getitem__.return_value = uuid_column

        # uuid already present
        product = model.Product(uuid='some_uuid')
        recorder.ensure_uuid(product)
        self.assertEqual(product.uuid, 'some_uuid')
        self.assertFalse(get_uuid.called)

        # no uuid yet, auto-generate
        uuid_column.foreign_keys = False
        get_uuid.return_value = 'another_uuid'
        product = model.Product()
        self.assertTrue(product.uuid is None)
        recorder.ensure_uuid(product)
        get_uuid.assert_called_once_with()
        self.assertEqual(product.uuid, 'another_uuid')

        # some heavy mocking for following tests
        uuid_column.foreign_keys = True
        remote_side = MagicMock(key='uuid')
        prop = MagicMock(__class__=orm.RelationshipProperty, key='foreign_thing')
        prop.remote_side.__len__.return_value = 1
        prop.remote_side.__iter__.return_value = [remote_side]
        object_mapper.return_value.iterate_properties.__iter__.return_value = [prop]
        
        # uuid fetched from existing foreign key object
        get_uuid.reset_mock()
        instance = Mock(uuid=None, foreign_thing=Mock(uuid='secondary_uuid'))
        recorder.ensure_uuid(instance)
        self.assertFalse(get_uuid.called)
        self.assertEqual(instance.uuid, 'secondary_uuid')

        # foreign key object doesn't exist; uuid generated as fallback
        get_uuid.return_value = 'fallback_uuid'
        instance = Mock(uuid=None, foreign_thing=None)
        recorder.ensure_uuid(instance)
        get_uuid.assert_called_once_with()
        self.assertEqual(instance.uuid, 'fallback_uuid')


class TestFunctionalChanges(DataTestCase):

    def setUp(self):
        super(TestFunctionalChanges, self).setUp()
        changes.record_changes(self.session)

    def test_add(self):
        product = model.Product()
        self.session.add(product)
        self.session.commit()

        self.assertEqual(self.session.query(model.Change).count(), 1)
        change = self.session.query(model.Change).one()
        self.assertEqual(change.class_name, 'Product')
        self.assertEqual(change.instance_uuid, product.uuid)
        self.assertFalse(change.deleted)

    def test_change(self):
        product = model.Product()
        self.session.add(product)
        self.session.commit()

        self.assertEqual(self.session.query(model.Change).count(), 1)
        self.session.query(model.Change).delete()
        self.assertEqual(self.session.query(model.Change).count(), 0)

        product.description = 'Acme Bricks'
        self.session.commit()

        self.assertEqual(self.session.query(model.Change).count(), 1)
        change = self.session.query(model.Change).one()
        self.assertEqual(change.class_name, 'Product')
        self.assertEqual(change.instance_uuid, product.uuid)
        self.assertFalse(change.deleted)

    def test_delete(self):
        product = model.Product()
        self.session.add(product)
        self.session.commit()

        self.assertEqual(self.session.query(model.Change).count(), 1)
        self.session.query(model.Change).delete()
        self.assertEqual(self.session.query(model.Change).count(), 0)

        self.session.delete(product)

        self.assertEqual(self.session.query(model.Change).count(), 1)
        change = self.session.query(model.Change).one()
        self.assertEqual(change.class_name, 'Product')
        self.assertEqual(change.instance_uuid, product.uuid)
        self.assertTrue(change.deleted)

    def test_orphan_change(self):
        department = model.Department()
        subdepartment = model.Subdepartment()
        department.subdepartments.append(subdepartment)
        self.session.add(department)
        self.session.commit()

        self.assertEqual(self.session.query(model.Change).count(), 2)
        change = self.session.query(model.Change).filter_by(class_name='Department').one()
        self.assertFalse(change.deleted)
        change = self.session.query(model.Change).filter_by(class_name='Subdepartment').one()
        self.assertFalse(change.deleted)

        self.session.query(model.Change).delete()
        self.assertEqual(self.session.query(model.Change).count(), 0)

        # Creating an orphaned Subdepartment, which should be recorded as a
        # *change* due to the cascade rules in effect.
        department.subdepartments.remove(subdepartment)
        self.session.commit()

        self.assertEqual(self.session.query(model.Change).count(), 2)
        change = self.session.query(model.Change).filter_by(class_name='Department').one()
        self.assertFalse(change.deleted)
        change = self.session.query(model.Change).filter_by(class_name='Subdepartment').one()
        self.assertFalse(change.deleted)
        self.assertEqual(self.session.query(model.Subdepartment).count(), 1)
    
    def test_orphan_delete(self):
        customer = model.Customer()
        group = model.CustomerGroup()
        customer.groups.append(group)
        self.session.add(customer)
        self.session.commit()

        self.assertEqual(self.session.query(model.Change).count(), 3)
        change = self.session.query(model.Change).filter_by(class_name='Customer').one()
        self.assertFalse(change.deleted)
        change = self.session.query(model.Change).filter_by(class_name='CustomerGroup').one()
        self.assertFalse(change.deleted)
        change = self.session.query(model.Change).filter_by(class_name='CustomerGroupAssignment').one()
        self.assertFalse(change.deleted)

        self.session.query(model.Change).delete()
        self.assertEqual(self.session.query(model.Change).count(), 0)

        # Creating an orphaned CustomerGroupAssociation, which should be
        # recorded as a *deletion* due to the cascade rules in effect.  Note
        # that the CustomerGroup is not technically an orphan and in fact is
        # not even changed.
        customer.groups.remove(group)
        self.session.commit()

        self.assertEqual(self.session.query(model.Change).count(), 2)
        change = self.session.query(model.Change).filter_by(class_name='Customer').one()
        self.assertFalse(change.deleted)
        change = self.session.query(model.Change).filter_by(class_name='CustomerGroupAssignment').one()
        self.assertTrue(change.deleted)
        self.assertEqual(self.session.query(model.CustomerGroupAssignment).count(), 0)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import os
import unittest
import ConfigParser

from mock import patch
from fixture import TempIO

from rattail import config


class TestParseBoolFunc(unittest.TestCase):

    def test_none(self):
        self.assertIsNone(config.parse_bool(None))

    def test_true(self):
        self.assertIs(config.parse_bool(True), True)

    def test_false(self):
        self.assertIs(config.parse_bool(False), False)

    def test_string(self):
        self.assertTrue(config.parse_bool('true'))
        self.assertTrue(config.parse_bool('yes'))
        self.assertTrue(config.parse_bool('on'))
        self.assertTrue(config.parse_bool('1'))

        self.assertFalse(config.parse_bool('false'))
        self.assertFalse(config.parse_bool('no'))
        self.assertFalse(config.parse_bool('off'))
        self.assertFalse(config.parse_bool('0'))


class TestParseListFunc(unittest.TestCase):

    def test_none(self):
        value = config.parse_list(None)
        self.assertEqual(len(value), 0)

    def test_single_value(self):
        value = config.parse_list(u'foo')
        self.assertEqual(len(value), 1)
        self.assertEqual(value[0], u'foo')

    def test_single_value_padded_by_spaces(self):
        value = config.parse_list(u'   foo   ')
        self.assertEqual(len(value), 1)
        self.assertEqual(value[0], u'foo')

    def test_slash_is_not_a_separator(self):
        value = config.parse_list(u'/dev/null')
        self.assertEqual(len(value), 1)
        self.assertEqual(value[0], u'/dev/null')

    def test_multiple_values_separated_by_whitespace(self):
        value = config.parse_list(u'foo bar baz')
        self.assertEqual(len(value), 3)
        self.assertEqual(value[0], u'foo')
        self.assertEqual(value[1], u'bar')
        self.assertEqual(value[2], u'baz')

    def test_multiple_values_separated_by_commas(self):
        value = config.parse_list(u'foo,bar,baz')
        self.assertEqual(len(value), 3)
        self.assertEqual(value[0], u'foo')
        self.assertEqual(value[1], u'bar')
        self.assertEqual(value[2], u'baz')

    def test_multiple_values_separated_by_whitespace_and_commas(self):
        value = config.parse_list(u'  foo,   bar   baz')
        self.assertEqual(len(value), 3)
        self.assertEqual(value[0], u'foo')
        self.assertEqual(value[1], u'bar')
        self.assertEqual(value[2], u'baz')

    def test_multiple_values_separated_by_whitespace_and_commas_with_some_quoting(self):
        value = config.parse_list("""
        foo
        "C:\\some path\\with spaces\\and, a comma",
        baz
        """)
        self.assertEqual(len(value), 3)
        self.assertEqual(value[0], u'foo')
        self.assertEqual(value[1], u'C:\\some path\\with spaces\\and, a comma')
        self.assertEqual(value[2], u'baz')


class TestRattailConfig(unittest.TestCase):

    def setUp(self):
        self.tempio = TempIO()

    def tearDown(self):
        del self.tempio

    def setup_files(self):
        self.site_path = self.tempio.putfile('site.conf', """
[rattail]
        """)

        self.host_path = self.tempio.putfile('host.conf', """
[rattail.config]
include = "{}"
        """.format(self.site_path))

        self.app_path = self.tempio.putfile('app.conf', """
[rattail.config]
include = "{}"
        """.format(self.host_path))

        self.custom_path = self.tempio.putfile('custom.conf', """
[rattail.config]
include = "%(here)s/app.conf"
        """)

    def test_init_defaults(self):
        cfg = config.RattailConfig()
        self.assertEqual(cfg.files_requested, [])
        self.assertEqual(cfg.files_read, [])
        self.assertIsNone(cfg._session_factory)

    def test_init_params(self):
        self.setup_files()

        # files
        cfg = config.RattailConfig()
        self.assertEqual(cfg.files_requested, [])
        self.assertEqual(cfg.files_read, [])
        cfg = config.RattailConfig(files=[self.site_path])
        self.assertEqual(cfg.files_requested, [self.site_path])
        self.assertEqual(cfg.files_read, [self.site_path])

        # usedb
        cfg = config.RattailConfig()
        self.assertFalse(cfg.usedb)
        cfg = config.RattailConfig(usedb=True)
        self.assertTrue(cfg.usedb)

        # preferdb
        cfg = config.RattailConfig()
        self.assertFalse(cfg.preferdb)
        cfg = config.RattailConfig(preferdb=True)
        self.assertTrue(cfg.preferdb)

    def test_read_file_with_recurse(self):
        self.setup_files()
        cfg = config.RattailConfig()
        cfg.read_file(self.custom_path, recurse=True)
        self.assertEqual(cfg.files_requested, [self.custom_path, self.app_path, self.host_path, self.site_path])
        self.assertEqual(cfg.files_read, [self.site_path, self.host_path, self.app_path, self.custom_path])

    def test_read_file_once_only(self):
        self.setup_files()

        another_path = self.tempio.putfile('another.conf', """
[rattail.config]
include = "{custom}" "{site}" "{app}" "{site}" "{custom}"
        """.format(custom=self.custom_path, app=self.app_path, site=self.site_path))

        cfg = config.RattailConfig()
        cfg.read_file(another_path, recurse=True)
        self.assertEqual(cfg.files_requested, [another_path, self.custom_path, self.app_path, self.host_path, self.site_path])
        self.assertEqual(cfg.files_read, [self.site_path, self.host_path, self.app_path, self.custom_path, another_path])

    def test_read_file_skip_missing(self):
        self.setup_files()
        bogus_path = '/tmp/does-not/exist'
        self.assertFalse(os.path.exists(bogus_path))

        another_path = self.tempio.putfile('another.conf', """
[rattail.config]
include = "{bogus}" "{app}" "{bogus}" "{site}"
        """.format(bogus=bogus_path, app=self.app_path, site=self.site_path))

        cfg = config.RattailConfig()
        cfg.read_file(another_path, recurse=True)
        self.assertEqual(cfg.files_requested, [another_path, bogus_path, self.app_path, self.host_path, self.site_path])
        self.assertEqual(cfg.files_read, [self.site_path, self.host_path, self.app_path, another_path])

    @patch('rattail.config.logging.config.fileConfig')
    def test_configure_logging(self, fileConfig):
        cfg = config.RattailConfig()

        # logging not configured by default
        cfg.configure_logging()
        self.assertFalse(fileConfig.called)

        # but config option can enable it
        cfg.set('rattail.config', 'configure_logging', 'true')
        cfg.configure_logging()
        self.assertEqual(fileConfig.call_count, 1)
        fileConfig.reset_mock()

        # invalid logging config is ignored
        fileConfig.side_effect = ConfigParser.NoSectionError('loggers')
        cfg.configure_logging()
        self.assertEqual(fileConfig.call_count, 1)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database Utilities
"""

from __future__ import unicode_literals, absolute_import

import re
import pprint
import logging

from sqlalchemy import orm
from formencode import Invalid
from formencode.validators import Email

# TODO: Deprecate/remove these imports.
from rattail.db.config import engine_from_config, get_engines, get_default_engine, configure_session


log = logging.getLogger(__name__)


class QuerySequence(object):
    """
    Simple wrapper for a SQLAlchemy (or Django, or other?) query, to make it
    sort of behave like a normal sequence, as much as needed to e.g. make an
    importer happy.
    """

    def __init__(self, query):
        self.query = query

    def __len__(self):
        return self.query.count()

    def __iter__(self):
        return iter(self.query)


def maxlen(attr):
    """
    Return the maximum length for the given attribute.
    """
    if len(attr.property.columns) == 1:
        type_ = attr.property.columns[0].type
        return getattr(type_, 'length', None)


def make_topo_sortkey(model, metadata=None):
    """
    Returns a function suitable for use as a ``key`` kwarg to a standard Python
    sorting call.  This key function will expect a single class mapper and
    return a sequence number associated with that model.  The sequence is
    determined by SQLAlchemy's topological table sorting.
    """
    if metadata is None:
        metadata = model.Base.metadata

    tables = {}
    for i, table in enumerate(metadata.sorted_tables, 1):
        tables[table.name] = i

    log.debug("topo sortkeys for '{}' will be:\n{}".format(model.__name__, pprint.pformat(
        [(i, name) for name, i in sorted(tables.iteritems(), key=lambda t: t[1])])))

    def sortkey(name):
        mapper = orm.class_mapper(getattr(model, name))
        return tuple(tables[t.name] for t in mapper.tables)

    return sortkey


def normalize_full_name(first_name, last_name):
    """
    Normalize the given first and last name to a "full" name value.  The
    fallback return value is an empty string.
    """
    first_name = (first_name or '').strip()
    last_name = (last_name or '').strip()
    if first_name and last_name:
        return "{} {}".format(first_name, last_name)
    if first_name:
        return first_name
    if last_name:
        return last_name
    return ''


def normalize_phone_number(number):
    """
    Normalize a phone number to digits only.
    """
    if number is not None:
        return re.sub(r'\D', '', number)


def format_phone_number(number):
    """
    Returns a phone number in ``(XXX) XXX-XXXX`` format if possible; otherwise
    returns the argument unaltered.
    """
    original, number = number, normalize_phone_number(number)
    if number and len(number) == 10:
        return '({}) {}-{}'.format(number[:3], number[3:6], number[6:])
    return original


class EmailValidator(object):
    """
    Simple validator, used to ensure an email address matches the general
    expected pattern.

    This leverages the FormEncode library for general address validation.  It
    can optionally use DNS lookups to confirm if the domain portion of the
    email address is valid.
    """

    def __init__(self, resolve_domain=False):
        self.validator = Email(resolve_domain=resolve_domain)

    def validate(self, address):
        if address:
            try:
                return self.validator.to_python(address)
            except Invalid:
                pass


def validate_email_address(address, resolve_domain=False):
    """
    Validate a single email address.
    """
    validator = EmailValidator(resolve_domain=resolve_domain)
    return validator.validate(address)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Products
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.ext.associationproxy import association_proxy

from rattail import enum
from .core import Base, uuid_column, GPCType, getset_factory
from .org import Department, Subdepartment, Category
from .vendors import Vendor


class Brand(Base):
    """
    Represents a brand or similar product line.
    """
    __tablename__ = 'brand'
    __versioned__ = {}

    uuid = uuid_column()
    name = sa.Column(sa.String(length=100))

    def __unicode__(self):
        return unicode(self.name or '')


class Tax(Base):
    """
    Represents a sales tax rate to be applied to products.
    """
    __tablename__ = 'tax'
    __table_args__ = (
        sa.UniqueConstraint('code', name='tax_uq_code'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    code = sa.Column(sa.Integer(), nullable=False)
    description = sa.Column(sa.String(length=255), nullable=True)
    rate = sa.Column(sa.Numeric(precision=7, scale=5), nullable=False)

    def __unicode__(self):
        return "{0} ({1:0.2f} %)".format(self.description, self.rate)


class Product(Base):
    """
    Represents a product for sale and/or purchase.
    """
    __tablename__ = 'product'
    __table_args__ = (
        sa.ForeignKeyConstraint(['department_uuid'], ['department.uuid'], name='product_fk_department'),
        sa.ForeignKeyConstraint(['subdepartment_uuid'], ['subdepartment.uuid'], name='product_fk_subdepartment'),
        sa.ForeignKeyConstraint(['category_uuid'], ['category.uuid'], name='product_fk_category'),
        sa.ForeignKeyConstraint(['family_uuid'], ['family.uuid'], name='product_fk_family'),
        sa.ForeignKeyConstraint(['brand_uuid'], ['brand.uuid'], name='product_fk_brand'),
        sa.ForeignKeyConstraint(['report_code_uuid'], ['report_code.uuid'], name='product_fk_report_code'),
        sa.ForeignKeyConstraint(['deposit_link_uuid'], ['deposit_link.uuid'], name='product_fk_deposit_link'),
        sa.ForeignKeyConstraint(['tax_uuid'], ['tax.uuid'], name='product_fk_tax'),
        sa.ForeignKeyConstraint(['regular_price_uuid'], ['product_price.uuid'], name='product_fk_regular_price', use_alter=True),
        sa.ForeignKeyConstraint(['current_price_uuid'], ['product_price.uuid'], name='product_fk_current_price', use_alter=True),
        sa.Index('product_ix_upc', 'upc'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    upc = sa.Column(GPCType())
    department_uuid = sa.Column(sa.String(length=32))
    subdepartment_uuid = sa.Column(sa.String(length=32))
    category_uuid = sa.Column(sa.String(length=32))
    family_uuid = sa.Column(sa.String(length=32))

    report_code_uuid = sa.Column(sa.String(length=32), nullable=True, doc=u"""\
UUID of the product's report code, if any.
""")

    deposit_link_uuid = sa.Column(sa.String(length=32), nullable=True, doc="""
UUID of the product's deposit link, if any.
""")

    deposit_link = orm.relationship('DepositLink', doc="""
Reference to the :class:`DepositLink` instance with which the product
associates, if any.
""")

    tax_uuid = sa.Column(sa.String(length=32), nullable=True, doc="""
UUID of the product's tax, if any.
""")

    tax = orm.relationship(Tax, doc="""
Reference to the :class:`Tax` instance with which the product associates, if
any.
""")

    brand_uuid = sa.Column(sa.String(length=32))
    description = sa.Column(sa.String(length=60))
    description2 = sa.Column(sa.String(length=60))
    size = sa.Column(sa.String(length=30))

    unit_size = sa.Column(sa.Numeric(precision=8, scale=3), nullable=True, doc="""
Unit size for product, as decimal.  This refers to the numeric size of a unit
of the product, in terms of the :attr:`unit_of_measure`, and may be used for
smarter price comparisons etc.
""")

    unit_of_measure = sa.Column(sa.String(length=4), nullable=False,
                                default=enum.UNIT_OF_MEASURE_NONE, doc="""
Code indicating the unit of measure for the product.  Value should be one of
the keys of the ``rattail.enum.UNIT_OF_MEASURE`` dictionary.
""")

    weighed = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
Whether or not the product must be weighed to determine its final price.
""")

    case_pack = sa.Column(sa.Integer(), nullable=True, doc="""
Pack size for the product, i.e. how many units are in a case.
""")

    organic = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
Whether the item is organic.
""")

    not_for_sale = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
Flag to indicate items which are not available for sale.
""")

    deleted = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
Flag to indicate items which have been deleted.  Obviously this is implies
"false" deletion, where the record is actually kept on file.  Whether or not
you use this is up to you.
""")

    regular_price_uuid = sa.Column(sa.String(length=32))
    current_price_uuid = sa.Column(sa.String(length=32))

    discountable = sa.Column(sa.Boolean(), nullable=False, default=True, doc="""
Whether or not the product may be discounted in any way.
""")

    special_order = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
Whether the product is available for special order.
""")

    food_stampable = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating whether food stamps are a valid form of payment for the item.
    """)

    tax1 = sa.Column(sa.Boolean(), nullable=True, doc="""
    (hack?) Flag indicating whether 'tax 1' applies to the item.
    """)

    tax2 = sa.Column(sa.Boolean(), nullable=True, doc="""
    (hack?) Flag indicating whether 'tax 2' applies to the item.
    """)

    tax3 = sa.Column(sa.Boolean(), nullable=True, doc="""
    (hack?) Flag indicating whether 'tax 3' applies to the item.
    """)

    last_sold = sa.Column(sa.DateTime(), nullable=True, doc="""
UTC timestamp of the product's last sale event.
""")

    department = orm.relationship(Department, order_by=Department.name)

    subdepartment = orm.relationship(
        Subdepartment,
        order_by=Subdepartment.name,
        backref=orm.backref('products', cascade='all'))

    category = orm.relationship(
        Category, back_populates='products')

    brand = orm.relationship(Brand)

    family = orm.relationship(
        u'Family', back_populates=u'products', doc=u"""\
Reference to the :class:`Family` instance with which the product associates, if
any.
""")

    report_code = orm.relationship(
        u'ReportCode', back_populates=u'products', doc=u"""\
Reference to the :class:`ReportCode` instance with which the product
associates, if any.
""")

    def __unicode__(self):
        return self.full_description

    @property
    def full_description(self):
        """
        Convenience attribute which returns a more complete description.

        Most notably, this includes the brand name and product size.
        """
        fields = [
            (self.brand.name or '') if self.brand else '',
            self.description or '',
            self.size or '']
        fields = [f.strip() for f in fields if f.strip()]
        return ' '.join(fields)

    @property
    def pretty_upc(self):
        """
        Product's UPC as a somewhat human-readable string.
        """
        if self.upc is None:
            return None
        return self.upc.pretty()

    def cost_for_vendor(self, vendor):
        """
        Of the product's cost records, return the first one associated with the
        given vendor.
        """
        for cost in self.costs:
            if cost.vendor is vendor:
                return cost


Category.products = orm.relationship(
    Product, back_populates='category')


class ProductCode(Base):
    """
    Represents an arbitrary "code" for a product.
    """
    __tablename__ = 'product_code'
    __table_args__ = (
        sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'], name='product_code_fk_product'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    product_uuid = sa.Column(sa.String(length=32), nullable=False)
    ordinal = sa.Column(sa.Integer(), nullable=False)
    code = sa.Column(sa.String(length=20))

    def __unicode__(self):
        return unicode(self.code or '')


Product._codes = orm.relationship(
    ProductCode, backref='product',
    collection_class=ordering_list('ordinal', count_from=1),
    order_by=ProductCode.ordinal,
    cascade='save-update, merge, delete, delete-orphan')

Product.codes = association_proxy(
    '_codes', 'code',
    getset_factory=getset_factory,
    creator=lambda c: ProductCode(code=c))

Product._code = orm.relationship(
    ProductCode,
    primaryjoin=sa.and_(
        ProductCode.product_uuid == Product.uuid,
        ProductCode.ordinal == 1,
        ),
    uselist=False,
    viewonly=True)

Product.code = association_proxy(
    '_code', 'code',
    getset_factory=getset_factory)


class ProductCost(Base):
    """
    Represents a source from which a product may be obtained via purchase.
    """
    __tablename__ = 'product_cost'
    __table_args__ = (
        sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'], name='product_cost_fk_product'),
        sa.ForeignKeyConstraint(['vendor_uuid'], ['vendor.uuid'], name='product_cost_fk_vendor'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    product_uuid = sa.Column(sa.String(length=32), nullable=False)
    vendor_uuid = sa.Column(sa.String(length=32), nullable=False)
    preference = sa.Column(sa.Integer(), nullable=False)
    code = sa.Column(sa.String(length=20))

    case_size = sa.Column(sa.Integer())
    case_cost = sa.Column(sa.Numeric(precision=9, scale=5))
    pack_size = sa.Column(sa.Integer())
    pack_cost = sa.Column(sa.Numeric(precision=9, scale=5))
    unit_cost = sa.Column(sa.Numeric(precision=9, scale=5))
    effective = sa.Column(sa.DateTime())

    vendor = orm.relationship(
        Vendor,
        backref=orm.backref('product_costs', cascade='all'))


Product.costs = orm.relationship(
    ProductCost, backref='product',
    collection_class=ordering_list('preference', count_from=1),
    order_by=ProductCost.preference,
    cascade='save-update, merge, delete, delete-orphan')

Product.cost = orm.relationship(
    ProductCost,
    primaryjoin=sa.and_(
        ProductCost.product_uuid == Product.uuid,
        ProductCost.preference == 1,
        ),
    uselist=False,
    viewonly=True)

Product.vendor = orm.relationship(
    Vendor,
    secondary=ProductCost.__table__,
    primaryjoin=sa.and_(
        ProductCost.product_uuid == Product.uuid,
        ProductCost.preference == 1,
        ),
    secondaryjoin=Vendor.uuid == ProductCost.vendor_uuid,
    uselist=False,
    viewonly=True)


class ProductPrice(Base):
    """
    Represents a price for a product.
    """
    __tablename__ = 'product_price'
    __table_args__ = (
        sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'], name='product_price_fk_product'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    product_uuid = sa.Column(sa.String(length=32), nullable=False)
    type = sa.Column(sa.Integer())
    level = sa.Column(sa.Integer())
    starts = sa.Column(sa.DateTime())
    ends = sa.Column(sa.DateTime())
    price = sa.Column(sa.Numeric(precision=8, scale=3))
    multiple = sa.Column(sa.Integer())
    pack_price = sa.Column(sa.Numeric(precision=8, scale=3))
    pack_multiple = sa.Column(sa.Integer())


Product.prices = orm.relationship(
    ProductPrice, backref='product',
    primaryjoin=ProductPrice.product_uuid == Product.uuid,
    cascade='save-update, merge, delete, delete-orphan')

Product.regular_price = orm.relationship(
    ProductPrice,
    primaryjoin=Product.regular_price_uuid == ProductPrice.uuid,
    lazy='joined',
    post_update=True)

Product.current_price = orm.relationship(
    ProductPrice,
    primaryjoin=Product.current_price_uuid == ProductPrice.uuid,
    lazy='joined',
    post_update=True)



# -*- coding: utf-8 -*-
"""add MailingAddress

Revision ID: 3df3144cec58
Revises: 4da3fe4a368f
Create Date: 2015-12-17 18:47:37.364765

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '3df3144cec58'
down_revision = u'4da3fe4a368f'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # address
    op.create_table('address',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('parent_type', sa.String(length=20), nullable=False),
                    sa.Column('parent_uuid', sa.String(length=32), nullable=False),
                    sa.Column('preference', sa.Integer(), nullable=False),
                    sa.Column('type', sa.String(length=15), nullable=True),
                    sa.Column('street', sa.String(length=100), nullable=True),
                    sa.Column('street2', sa.String(length=100), nullable=True),
                    sa.Column('city', sa.String(length=60), nullable=True),
                    sa.Column('state', sa.String(length=2), nullable=True),
                    sa.Column('zipcode', sa.String(length=10), nullable=True),
                    sa.Column('invalid', sa.Boolean(), nullable=True),
                    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('address_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('parent_type', sa.String(length=20), autoincrement=False, nullable=True),
                    sa.Column('parent_uuid', sa.String(length=32), autoincrement=False, nullable=True),
                    sa.Column('preference', sa.Integer(), autoincrement=False, nullable=True),
                    sa.Column('type', sa.String(length=15), autoincrement=False, nullable=True),
                    sa.Column('street', sa.String(length=100), nullable=True),
                    sa.Column('street2', sa.String(length=100), nullable=True),
                    sa.Column('city', sa.String(length=60), nullable=True),
                    sa.Column('state', sa.String(length=2), nullable=True),
                    sa.Column('zipcode', sa.String(length=10), nullable=True),
                    sa.Column('invalid', sa.Boolean(), nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_address_version_end_transaction_id'), 'address_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_address_version_operation_type'), 'address_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_address_version_transaction_id'), 'address_version', ['transaction_id'], unique=False)


def downgrade():

    # address
    op.drop_index(op.f('ix_address_version_transaction_id'), table_name='address_version')
    op.drop_index(op.f('ix_address_version_operation_type'), table_name='address_version')
    op.drop_index(op.f('ix_address_version_end_transaction_id'), table_name='address_version')
    op.drop_table('address_version')
    op.drop_table('address')



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import Queue
from unittest import TestCase

from fixture import TempIO

from rattail.config import make_config
from rattail.filemon import util
from rattail.filemon.config import Profile


class TestQueueExisting(TestCase):

    def setUp(self):
        self.tmp = TempIO()
        self.config = make_config([])
        self.config.set(u'rattail.filemon', u'monitor', u'foo')
        self.config.set(u'rattail.filemon', u'foo.dirs', self.tmp)
        self.config.set(u'rattail.filemon', u'foo.actions', u'noop')
        self.config.set(u'rattail.filemon', u'foo.action.noop.func', u'rattail.filemon.actions:noop')
        self.profile = Profile(self.config, u'foo')
        self.profile.queue = Queue.Queue()

    def test_nothing_queued_if_no_files_exist(self):
        util.queue_existing(self.profile, self.tmp)
        self.assertTrue(self.profile.queue.empty())

    def test_normal_files_are_queued_but_not_folders(self):
        self.tmp.putfile(u'file', u'')
        self.tmp.mkdir(u'folder')
        util.queue_existing(self.profile, self.tmp)
        self.assertEqual(self.profile.queue.qsize(), 1)
        self.assertEqual(self.profile.queue.get_nowait(), self.tmp.join(u'file'))
        self.assertTrue(self.profile.queue.empty())

    def test_if_profile_watches_locks_then_normal_files_are_queued_but_not_lock_files(self):
        self.profile.watch_locks = True
        self.tmp.putfile(u'file1.lock', u'')
        self.tmp.putfile(u'file2', u'')
        util.queue_existing(self.profile, self.tmp)
        self.assertEqual(self.profile.queue.qsize(), 1)
        self.assertEqual(self.profile.queue.get_nowait(), self.tmp.join(u'file2'))
        self.assertTrue(self.profile.queue.empty())



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from unittest import TestCase
from mock import patch, call, Mock, DEFAULT

from sqlalchemy.exc import OperationalError

from . import SyncTestCase
from rattail.db import sync
from rattail.db import model
from rattail.config import make_config


class SynchronizerTests(SyncTestCase):

    def test_init(self):
        synchronizer = sync.Synchronizer(self.local_engine, self.remote_engines)
        self.assertTrue(synchronizer.local_engine is self.local_engine)
        self.assertTrue(synchronizer.remote_engines is self.remote_engines)

    def test_loop(self):

        class FakeOperationalError(OperationalError):
            def __init__(self, connection_invalidated):
                self.connection_invalidated = connection_invalidated

        synchronizer = sync.Synchronizer(self.local_engine, self.remote_engines)
        with patch.object(synchronizer, 'sleep') as sleep:
            with patch.object(synchronizer, 'synchronize') as synchronize:

                synchronize.side_effect = [1, 2, 3, FakeOperationalError(True),
                                                   5, 6, 7, FakeOperationalError(False)]
                self.assertRaises(FakeOperationalError, synchronizer.loop)
                self.assertEqual(synchronize.call_count, 8)
                self.assertEqual(sleep.call_args_list, [
                        call(3), call(3), call(3), call(5), call(3),
                        call(3), call(3), call(3)])

    # def test_synchronize(self):
    #     synchronizer = sync.Synchronizer(self.local_engine, self.remote_engines)

    #     with patch.object(synchronizer, 'synchronize_changes') as synchronize_changes:

    #         # no changes
    #         synchronizer.synchronize()
    #         self.assertFalse(synchronize_changes.called)

    #         # some changes
    #         local_session = self.Session(bind=self.local_engine)
    #         product = model.Product()
    #         local_session.add(product)
    #         local_session.flush()
    #         local_session.add(model.Change(class_name='Product', uuid=product.uuid, deleted=False))
    #         product = model.Product()
    #         local_session.add(product)
    #         local_session.flush()
    #         local_session.add(model.Change(class_name='Product', uuid=product.uuid, deleted=False))
    #         local_session.commit()
    #         synchronizer.synchronize()
    #         self.assertEqual(synchronize_changes.call_count, 1)
    #         # call_args is a tuple of (args, kwargs) - first element of args should be our 2 changes
    #         self.assertEqual(len(synchronize_changes.call_args[0][0]), 2)
    #         self.assertTrue(isinstance(synchronize_changes.call_args[0][0][0], model.Change))

    # def test_synchronize_changes(self):
    #     synchronizer = sync.Synchronizer(self.local_engine, self.remote_engines)

    #     local_session = self.Session(bind=self.local_engine)
    #     remote_sessions = {
    #         'one': self.Session(bind=self.remote_engines['one']),
    #         'two': self.Session(bind=self.remote_engines['two']),
    #         }

    #     # no changes; nothing should happen but make sure nothing blows up also
    #     local_changes = []
    #     synchronizer.synchronize_changes(local_changes, local_session, remote_sessions)

    #     # add a product, with change
    #     product = model.Product()
    #     local_session.add(product)
    #     local_session.flush()
    #     change = model.Change(class_name='Product', uuid=product.uuid, deleted=False)
    #     local_session.add(change)
    #     local_session.flush()
    #     self.assertEqual(local_session.query(model.Product).count(), 1)
    #     self.assertEqual(local_session.query(model.Change).count(), 1)

    #     # remote sessions don't have the product yet
    #     self.assertEqual(remote_sessions['one'].query(model.Product).count(), 0)
    #     self.assertEqual(remote_sessions['two'].query(model.Product).count(), 0)

    #     # sync the change
    #     synchronizer.synchronize_changes([change], local_session, remote_sessions)
    #     self.assertEqual(local_session.query(model.Product).count(), 1)
    #     self.assertEqual(local_session.query(model.Change).count(), 0)

    #     # remote session 'one' has the product
    #     self.assertEqual(remote_sessions['one'].query(model.Product).count(), 1)
    #     remote_product_1 = remote_sessions['one'].query(model.Product).one()
    #     self.assertEqual(remote_product_1.uuid, product.uuid)

    #     # remote session 'two' has the product
    #     self.assertEqual(remote_sessions['two'].query(model.Product).count(), 1)
    #     remote_product_2 = remote_sessions['two'].query(model.Product).one()
    #     self.assertEqual(remote_product_2.uuid, product.uuid)

    #     # delete the product (new change)
    #     local_session.delete(product)
    #     change = model.Change(class_name='Product', uuid=product.uuid, deleted=True)
    #     local_session.add(change)
    #     local_session.flush()
    #     self.assertEqual(local_session.query(model.Product).count(), 0)
    #     self.assertEqual(local_session.query(model.Change).count(), 1)

    #     # sync the change
    #     synchronizer.synchronize_changes([change], local_session, remote_sessions)
    #     self.assertEqual(local_session.query(model.Change).count(), 0)

    #     # remote sessions no longer have the product
    #     self.assertEqual(remote_sessions['one'].query(model.Product).count(), 0)
    #     self.assertEqual(remote_sessions['two'].query(model.Product).count(), 0)

    def test_dependency_sort(self):
        synchronizer = sync.Synchronizer(self.local_engine, self.remote_engines)

        # Product depends on Department, so Department should come first.
        self.assertEqual(synchronizer.dependency_sort('Department', 'Product'), -1)
        self.assertEqual(synchronizer.dependency_sort('Product', 'Department'), 1)

        # Product has dependencies (e.g. Department), so should come after
        # e.g. Store even though there is no direct connection, since Store has
        # no dependencies.
        self.assertEqual(synchronizer.dependency_sort('Store', 'Product'), -1)
        self.assertEqual(synchronizer.dependency_sort('Product', 'Store'), 1)

        # Sometimes the tie can't be broken...
        self.assertEqual(synchronizer.dependency_sort('Store', 'CustomerGroup'), 0)
        self.assertEqual(synchronizer.dependency_sort('Product', 'Product'), 0)

    def test_merge_instance(self):

        class FakeClass(object):
            pass

        synchronizer = sync.Synchronizer(self.local_engine, self.remote_engines)
        session = Mock()
        instance = FakeClass()

        self.assertFalse(hasattr(synchronizer, 'merge_FakeClass'))
        synchronizer.merge_instance(session, instance)
        session.merge.assert_called_once_with(instance)

        synchronizer.merge_FakeClass = Mock()
        synchronizer.merge_instance(session, instance)
        synchronizer.merge_FakeClass.assert_called_once_with(session, instance)

    def test_merge_Product(self):
        synchronizer = sync.Synchronizer(self.local_engine, self.remote_engines)

        # no prices
        local_session = self.Session(bind=self.local_engine)
        remote_session = self.Session(bind=self.remote_engines['one'])
        source_product = model.Product()
        local_session.add(source_product)
        local_session.flush()
        self.assertTrue(source_product.regular_price_uuid is None)
        self.assertTrue(source_product.regular_price is None)
        self.assertTrue(source_product.current_price_uuid is None)
        self.assertTrue(source_product.current_price is None)
        target_product = synchronizer.merge_Product(remote_session, source_product)
        self.assertFalse(target_product is None)
        self.assertFalse(source_product is target_product)
        self.assertEqual(source_product.uuid, target_product.uuid)
        self.assertTrue(target_product.regular_price_uuid is None)
        self.assertTrue(target_product.regular_price is None)
        self.assertTrue(target_product.current_price_uuid is None)
        self.assertTrue(target_product.current_price is None)
        local_session.rollback()
        local_session.close()
        remote_session.rollback()
        remote_session.close()

        # regular price
        local_session = self.Session(bind=self.local_engine)
        remote_session = self.Session(bind=self.remote_engines['one'])
        source_product = model.Product()
        regular_price = model.ProductPrice()
        source_product.prices.append(regular_price)
        source_product.regular_price = regular_price
        local_session.add(source_product)
        local_session.flush()
        self.assertFalse(source_product.regular_price_uuid is None)
        self.assertFalse(source_product.regular_price is None)
        target_product = synchronizer.merge_Product(remote_session, source_product)
        self.assertEqual(target_product.regular_price_uuid, source_product.regular_price_uuid)
        self.assertFalse(target_product.regular_price is None)
        local_session.rollback()
        local_session.close()
        remote_session.rollback()
        remote_session.close()

        # current price
        local_session = self.Session(bind=self.local_engine)
        remote_session = self.Session(bind=self.remote_engines['one'])
        source_product = model.Product()
        current_price = model.ProductPrice()
        source_product.prices.append(current_price)
        source_product.current_price = current_price
        local_session.add(source_product)
        local_session.flush()
        self.assertFalse(source_product.current_price_uuid is None)
        self.assertFalse(source_product.current_price is None)
        target_product = synchronizer.merge_Product(remote_session, source_product)
        self.assertEqual(target_product.current_price_uuid, source_product.current_price_uuid)
        self.assertFalse(target_product.current_price is None)
        local_session.rollback()
        local_session.close()
        remote_session.rollback()
        remote_session.close()

    def test_delete_instance(self):

        class FakeClass(object):
            pass

        synchronizer = sync.Synchronizer(self.local_engine, self.remote_engines)
        session = Mock()
        instance = FakeClass()

        self.assertFalse(hasattr(synchronizer, 'delete_FakeClass'))
        synchronizer.delete_instance(session, instance)
        session.delete.assert_called_once_with(instance)

        synchronizer.delete_FakeClass = Mock()
        synchronizer.delete_instance(session, instance)
        synchronizer.delete_FakeClass.assert_called_once_with(session, instance)

    def test_delete_Department(self):
        synchronizer = sync.Synchronizer(self.local_engine, self.remote_engines)

        session = self.Session(bind=self.local_engine)
        department = model.Department()
        department.subdepartments.append(model.Subdepartment())
        session.add(department)
        session.flush()
        self.assertEqual(session.query(model.Subdepartment).count(), 1)
        subdepartment = session.query(model.Subdepartment).one()
        self.assertEqual(subdepartment.department_uuid, department.uuid)
        synchronizer.delete_Department(session, department)
        self.assertEqual(session.query(model.Subdepartment).count(), 1)
        subdepartment = session.query(model.Subdepartment).one()
        self.assertTrue(subdepartment.department_uuid is None)
        session.rollback()
        session.close()

        session = self.Session(bind=self.local_engine)
        department = model.Department()
        product = model.Product(department=department)
        session.add(product)
        session.flush()
        product = session.query(model.Product).one()
        self.assertEqual(product.department_uuid, department.uuid)
        synchronizer.delete_Department(session, department)
        product = session.query(model.Product).one()
        self.assertTrue(product.department_uuid is None)
        session.rollback()
        session.close()

    def test_delete_Subdepartment(self):
        synchronizer = sync.Synchronizer(self.local_engine, self.remote_engines)

        session = self.Session(bind=self.local_engine)
        subdepartment = model.Subdepartment()
        product = model.Product(subdepartment=subdepartment)
        session.add(product)
        session.flush()
        product = session.query(model.Product).one()
        self.assertEqual(product.subdepartment_uuid, subdepartment.uuid)
        synchronizer.delete_Subdepartment(session, subdepartment)
        product = session.query(model.Product).one()
        self.assertTrue(product.subdepartment_uuid is None)
        session.rollback()
        session.close()

    def test_delete_Family(self):
        synchronizer = sync.Synchronizer(self.local_engine, self.remote_engines)

        session = self.Session(bind=self.local_engine)
        family = model.Family()
        product = model.Product(family=family)
        session.add(product)
        session.flush()
        product = session.query(model.Product).one()
        self.assertEqual(product.family_uuid, family.uuid)
        synchronizer.delete_Family(session, family)
        product = session.query(model.Product).one()
        self.assertTrue(product.family_uuid is None)
        session.rollback()
        session.close()

    def test_delete_Vendor(self):
        synchronizer = sync.Synchronizer(self.local_engine, self.remote_engines)

        session = self.Session(bind=self.local_engine)
        vendor = model.Vendor()
        product = model.Product()
        product.costs.append(model.ProductCost(vendor=vendor))
        session.add(product)
        session.flush()
        cost = session.query(model.ProductCost).one()
        self.assertEqual(cost.vendor_uuid, vendor.uuid)
        synchronizer.delete_Vendor(session, vendor)
        self.assertEqual(session.query(model.ProductCost).count(), 0)
        session.rollback()
        session.close()

    def test_delete_CustomerGroup(self):
        synchronizer = sync.Synchronizer(self.local_engine, self.remote_engines)

        session = self.Session(bind=self.local_engine)
        group = model.CustomerGroup()
        customer = model.Customer()
        customer.groups.append(group)
        session.add(customer)
        session.flush()
        assignment = session.query(model.CustomerGroupAssignment).one()
        self.assertEqual(assignment.customer_uuid, customer.uuid)
        self.assertEqual(assignment.group_uuid, group.uuid)
        synchronizer.delete_CustomerGroup(session, group)
        self.assertEqual(session.query(model.CustomerGroupAssignment).count(), 0)
        session.rollback()
        session.close()


class ModuleTests(TestCase):

    def setUp(self):
        self.config = make_config([])

    def test_get_sync_engines(self):
        config = Mock()
        config.get.return_value = None
        self.assertTrue(sync.get_sync_engines(config) is None)

        with patch('rattail.db.sync.get_engines') as get_engines:
            get_engines.return_value = {
                'one': 'first',
                'two': 'second',
                'three': 'third',
                }
            config.get.return_value = 'one, two'
            self.assertEqual(sync.get_sync_engines(config), {'one': 'first', 'two': 'second'})
            get_engines.assert_called_once_with(config)

    @patch('rattail.db.sync.load_object')
    @patch('rattail.db.sync.Synchronizer')
    def test_synchronize_changes(self, Synchronizer, load_object):

        local_engine = Mock()
        remote_engines = Mock()

        # default synchronizer class
        sync.synchronize_changes(self.config, local_engine, remote_engines)
        Synchronizer.assert_called_once_with(local_engine, remote_engines)
        Synchronizer.return_value.loop.assert_called_once_with()

        # custom synchronizer class
        self.config.set('rattail.db', 'sync.synchronizer_class', 'some_class')
        sync.synchronize_changes(self.config, local_engine, remote_engines)
        load_object.return_value.assert_called_once_with(local_engine, remote_engines)
        load_object.return_value.return_value.loop.assert_called_once_with()



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Authentication & Authorization
"""

from __future__ import unicode_literals, absolute_import

import bcrypt

from rattail.db import model


def authenticate_user(session, username, password):
    """
    Attempt to authenticate a user.

    :returns: A :class:`.model.User` instance, if successful.  Otherwise
       returns ``None``.
    """
    user = session.query(model.User)\
        .filter_by(username=username)\
        .first()
    if user and user.active and user.password is not None:
        # Apparently bcrypt's hashpw() doesn't like Unicode?
        try:
            authenticated = (bcrypt.hashpw(password, user.salt) == user.password)
        except UnicodeEncodeError:
            authenticated = False
        if authenticated:
            return user
    return None


def set_user_password(user, password):
    """
    Set a user's password.
    """

    user.salt = bcrypt.gensalt()
    user.password = bcrypt.hashpw(password, user.salt)


def special_role(session, uuid, name):
    """
    Fetches, or creates, a "special" role.
    """

    role = session.query(model.Role).get(uuid)
    if not role:
        role = model.Role(uuid=uuid, name=name)
        session.add(role)
    return role


def administrator_role(session):
    """
    Returns the "Administrator" role.
    """

    return special_role(session, 'd937fa8a965611dfa0dd001143047286', 'Administrator')


def guest_role(session):
    """
    Returns the "Guest" role.
    """

    return special_role(session, 'f8a27c98965a11dfaff7001143047286', 'Guest')


def authenticated_role(session):
    """
    Returns the "Authenticated" role.
    """
    return special_role(session, 'b765a9cc331a11e6ac2a3ca9f40bc550', "Authenticated")


def grant_permission(role, permission):
    """
    Grant a permission to a role.
    """

    # TODO: Make this a `Role` method (or make `Role.permissions` a `set` so we
    # can do `role.permissions.add('some.perm')` ?).
    if permission not in role.permissions:
        role.permissions.append(permission)


def has_permission(session, principal, permission, include_guest=True, include_authenticated=True):
    """
    Determine if a principal has been granted a permission.

    :param session: A SQLAlchemy session instance.

    :param principal: May be either a :class:`.model.User` or
       :class:`.model.Role` instance.  It is also expected that this may
       sometimes be ``None``, in which case the "Guest" role will typically be
       assumed.

    :param permission: The full internal name of a permission,
       e.g. ``'users.create'``.

    :param include_guest: Whether or not the "Guest" role should be included
       when checking permissions.  If ``False``, then Guest's permissions will
       *not* be consulted.

    :param include_authenticated: Whether or not the "Authenticated" role
       should be included when checking permissions.

    Note that if no ``principal`` is provided, and ``include_guest`` is set to
    ``False``, then no checks will actually be done, and the return value will
    be ``False``.
    """

    if hasattr(principal, 'roles'):
        roles = list(principal.roles)
        if include_authenticated:
            roles.append(authenticated_role(session))
    elif principal is not None:
        roles = [principal]
    else:
        roles = []

    # Admin always has permission.
    admin = administrator_role(session)
    if admin in roles:
        return True

    if include_guest:
        roles.append(guest_role(session))
    for role in roles:
        for perm in role.permissions:
            if perm == permission:
                return True
    return False



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from unittest import TestCase

from mock import patch

from rattail.sil import columns
from rattail.sil.exceptions import SILColumnNotFound


class TestGetColumn(TestCase):

    def test_exception_raised_if_no_such_column_exists(self):
        with patch('rattail.sil.columns.supported_columns', new={}):
            self.assertRaises(SILColumnNotFound, columns.get_column, 'bogus-column')



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for User Messages
"""

from __future__ import unicode_literals, absolute_import

import datetime

import sqlalchemy as sa
from sqlalchemy import orm

from .core import Base, uuid_column
from .users import User


class Message(Base):
    """
    Represents a message, sent from one user to other user(s).
    """
    __tablename__ = 'message'
    __table_args__ = (
        sa.ForeignKeyConstraint(['sender_uuid'], ['user.uuid'], name='message_fk_sender'),
    )

    uuid = uuid_column()
    sender_uuid = sa.Column(sa.String(length=32), nullable=False)

    sender = orm.relationship(
        User, doc="""
        Reference to the user who sent the message.
        """,
        backref=orm.backref('sent_messages', cascade='all, delete-orphan', doc="""
        List of all messages which have ever been sent by the user.
        """))

    subject = sa.Column(sa.String(length=255), nullable=True, doc="""
    Subject for the message.
    """)

    body = sa.Column(sa.Text(), nullable=True, doc="""
    Body for the message.  This is assumed to be of type 'text/html'.
    """)

    sent = sa.Column(sa.DateTime(), nullable=False, default=datetime.datetime.utcnow, doc="""
    UTC timestamp when the message was sent.
    """)

    def __unicode__(self):
        return unicode(self.subject or '')

    def add_recipient(self, user, **kwargs):
        """
        Add the given user to the message's recipients list, unless it's
        already there.
        """
        if not self.has_recipient(user):
            kwargs['recipient'] = user
            self.recipients.append(MessageRecipient(**kwargs))

    def has_recipient(self, user):
        """
        Returns boolean indicating whether the given user is listed among the
        message's recipients.
        """
        for recipient in self.recipients:
            if recipient.recipient is user:
                return True
        return False


class MessageRecipient(Base):
    """
    Represents the combination of a single message and a single recipient.
    Also tracks status of the message for that recipient, i.e. whether it shows
    in their "inbox".
    """
    __tablename__ = 'message_recip'
    __table_args__ = (
        sa.ForeignKeyConstraint(['message_uuid'], ['message.uuid'], name='message_recip_fk_message'),
        sa.ForeignKeyConstraint(['recipient_uuid'], ['user.uuid'], name='message_recip_fk_recipient'),
        sa.UniqueConstraint('message_uuid', 'recipient_uuid', name='message_recip_uq_message_recipient'),
    )

    uuid = uuid_column()
    message_uuid = sa.Column(sa.String(length=32), nullable=False)
    recipient_uuid = sa.Column(sa.String(length=32), nullable=False)

    status = sa.Column(sa.Integer(), nullable=False, doc="""
    Status code for the message; used to indicate inbox vs. archive.
    """)

    message = orm.relationship(
        Message, doc="""
        Reference to the message which has been "received".
        """,
        backref=orm.backref('recipients', cascade='all, delete-orphan', doc="""
        List of recipients for the message.
        """,
        ))

    recipient = orm.relationship(
        User, doc="""
        Reference to the user who "received" the message.
        """)

    def __unicode__(self):
        return unicode(self.recipient)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Models for vendor invoice batches
"""

from __future__ import unicode_literals

import sqlalchemy as sa
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr

from rattail.db.model import Base, Vendor
from rattail.db.batch.model import FileBatchMixin, ProductBatchRowMixin


class VendorInvoice(FileBatchMixin, Base):
    """
    Vendor invoice, the source data file of which has been provided by a user,
    and may be further processed in some site-specific way.
    """
    __tablename__ = 'vendor_invoice'
    __batchrow_class__ = 'VendorInvoiceRow'

    @declared_attr
    def __table_args__(cls):
        return cls.__default_table_args__() + (
            sa.ForeignKeyConstraint(['vendor_uuid'], ['vendor.uuid'],
                                    name='vendor_invoice_fk_vendor'),
            )

    parser_key = sa.Column(sa.String(length=100), nullable=False, doc="""
The key of the parser used to parse the contents of the data file.
""")

    vendor_uuid = sa.Column(sa.String(length=32), nullable=False)

    vendor = relationship(Vendor, doc="""
Reference to the :class:`Vendor` to which the invoice pertains.
""")

    invoice_date = sa.Column(sa.Date(), nullable=True, doc="""
Invoice date, as determined by the invoice data file.
""")

    purchase_order_number = sa.Column(sa.String(length=20), nullable=True, doc="""
Purchase order number, e.g. for cross-reference with another system.  Custom
batch handlers may populate and leverage this field, but the default handler
does not.
""")

    def add_row(self, **kwargs):
        """
        Add a :class:`VendorInvoiceRow` to the invoice.  This primarily is for
        convenience when a sequence number is not known, e.g. when adding
        "dummy" rows to the invoice batch for products which exist on a
        purchase order but not in the invoice data file.
        """
        if 'sequence' not in kwargs:
            kwargs['sequence'] = (max([r.sequence for r in self.data_rows]) or 0) + 1
        row = VendorInvoiceRow(**kwargs)
        self.data_rows.append(row)
        return row


class VendorInvoiceRow(ProductBatchRowMixin, Base):
    """
    Row of data within a vendor invoice.
    """
    __tablename__ = 'vendor_invoice_row'
    __batch_class__ = VendorInvoice

    STATUS_OK = 1
    STATUS_NOT_IN_DB = 2
    STATUS_NOT_IN_PURCHASE = 3
    STATUS_NOT_IN_INVOICE = 4
    STATUS_DIFFERS_FROM_PURCHASE = 5
    STATUS_COST_NOT_IN_DB = 6

    STATUS = {
        STATUS_OK:                      "as expected",
        STATUS_NOT_IN_DB:               "product not found",
        STATUS_COST_NOT_IN_DB:          "product found but not cost",
        STATUS_NOT_IN_PURCHASE:         "present only in invoice",
        STATUS_NOT_IN_INVOICE:          "present only in PO",
        STATUS_DIFFERS_FROM_PURCHASE:   "invoice and PO differ",
        }

    vendor_code = sa.Column(sa.String(length=30), nullable=True, doc="""
Vendor's unique code for the product.  The meaning of this corresponds to that
of the :attr:`ProductCost.code` column.
""")

    case_quantity = sa.Column(sa.Integer(), nullable=False, default=1, doc="""
Number of units in a case of product.
""")

    ordered_cases = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
Number of cases of the product which were originally ordered from the vendor.
""")

    ordered_units = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
Number of units of the product which were originally ordered from the vendor.
""")

    shipped_cases = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
Number of cases of the product which were shipped by the vendor.
""")

    shipped_units = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
Number of units of the product which were shipped by the vendor.
""")

    case_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
Cost per case of the product.
""")

    unit_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
Cost per unit of the product.
""")

    total_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
Total cost for this product line item; should equate to the number of units
shipped multiplied by the unit cost.
""")

    line_number = sa.Column(sa.Integer(), nullable=True, doc="""
Line number of the purchase order with which this invoice row matches.  Custom
batch handlers may populate and leverage this field, but the default handler
does not.
""")



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

from rattail.db import model, auth
from rattail.importing import model as import_model
from rattail.tests import RattailTestCase


class TestAdminUser(RattailTestCase):

    def make_importer(self, **kwargs):
        kwargs.setdefault('session', self.session)
        return import_model.AdminUserImporter(**kwargs)

    def get_admin(self):
        return auth.administrator_role(self.session)

    def test_supported_fields(self):
        importer = import_model.UserImporter()
        standard_fields = importer.fields
        importer = self.make_importer()
        extra_fields = set(importer.fields) - set(standard_fields)
        self.assertEqual(len(extra_fields), 1)
        self.assertEqual(list(extra_fields)[0], 'admin')

    def test_normalize_local_object(self):
        importer = self.make_importer()
        user = model.User()
        data = importer.normalize_local_object(user)
        self.assertFalse(data['admin'])
        user.roles.append(self.get_admin())
        data = importer.normalize_local_object(user)
        self.assertTrue(data['admin'])

    def test_update_object(self):
        importer = self.make_importer(fields=['uuid', 'admin'])
        data = {'uuid': 'ccb1915419e511e6a3ad3ca9f40bc550'}
        user = model.User(**data)
        admin = self.get_admin()
        self.assertNotIn(admin, user.roles)

        data['admin'] = True
        importer.update_object(user, data)
        self.assertIn(admin, user.roles)

        data['admin'] = False
        importer.update_object(user, data)
        self.assertNotIn(admin, user.roles)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2014 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Core Fabric Library
"""

from __future__ import unicode_literals

import os
import tempfile

from fabric.api import sudo, run, local, env, settings, put as fab_put
from fabric.contrib.files import append, upload_template as fab_upload_template

from mako.template import Template


def get_debian_version():
    """
    Fetch the version of Debian running on the target system.
    """
    version = unicode(run('cat /etc/debian_version'))
    version = '.'.join(version.split('.')[:2])
    return float(version)


def mkdir(paths, owner='root:root', mode=None):
    """
    Recursively make one or more directories.
    """
    if isinstance(paths, basestring):
        paths = [paths]
    sudo('mkdir --parents {0}'.format(' '.join(paths)))
    sudo('chown {0} {1}'.format(owner, ' '.join(paths)))
    if mode is not None:
        sudo('chmod {0} {1}'.format(mode, ' '.join(paths)))


def make_system_user(name='rattail', home='/srv/rattail', uid=None, shell=None, alias=False):
    """
    Make a new system user account, with the given home folder and shell path.
    Optionally add a mail alias to root as well.
    """
    with settings(warn_only=True):
        result = sudo('getent passwd {0}'.format(name))
    if result.failed:
        uid = '--uid {0}'.format(uid) if uid else ''
        shell = '--shell {0}'.format(shell) if shell else ''
        sudo('adduser --system --home {0} --group {1} {2} {3}'.format(home, name, uid, shell))
    if alias:
        add_mail_alias(name)


def add_mail_alias(local_name, aliased_to='root'):
    append('/etc/aliases', '{}: {}'.format(local_name, aliased_to), use_sudo=True)
    sudo('newaliases')


def put(local_path, remote_path, owner='root:root', **kwargs):
    """
    Put a file on the server, and set its ownership.
    """
    if 'mode' not in kwargs:
        kwargs.setdefault('mirror_local_mode', True)
    kwargs['use_sudo'] = True
    fab_put(local_path, remote_path, **kwargs)
    sudo('chown {0} {1}'.format(owner, remote_path))


def upload_template(local_path, remote_path, owner='root:root', **kwargs):
    """
    Upload a template to the server, and set its ownership.
    """
    if 'mode' not in kwargs:
        kwargs.setdefault('mirror_local_mode', True)
    kwargs['use_sudo'] = True
    fab_upload_template(local_path, remote_path, **kwargs)
    sudo('chown {0} {1}'.format(owner, remote_path))


def upload_mako_template(local_path, remote_path, owner='root:root', **kwargs):
    """
    Render a local file as a Mako template, and upload the result to the server.
    """
    template = Template(filename=local_path)

    temp_dir = tempfile.mkdtemp(prefix='rattail.fablib.')
    temp_path = os.path.join(temp_dir, os.path.basename(local_path))
    with open(temp_path, 'wb') as f:
        f.write(template.render(env=env))
    os.chmod(temp_path, os.stat(local_path).st_mode)

    put(temp_path, remote_path, **kwargs)
    os.remove(temp_path)
    os.rmdir(temp_dir)


class Deployer(object):

    def __init__(self, deploy_path, last_segment='deploy'):
        if not os.path.isdir(deploy_path):
            deploy_path = os.path.abspath(os.path.join(os.path.dirname(deploy_path), last_segment))
        self.deploy_path = deploy_path

    def __call__(self, local_path, remote_path, **kwargs):
        self.deploy(local_path, remote_path, **kwargs)

    def deploy(self, local_path, remote_path, **kwargs):
        local_path = '{0}/{1}'.format(self.deploy_path, local_path)
        if local_path.endswith('.template'):
            upload_template(local_path, remote_path, context=env, **kwargs)
        elif local_path.endswith('.mako'):
            upload_mako_template(local_path, remote_path, **kwargs)
        else:
            put(local_path, remote_path, **kwargs)

    def sudoers(self, local_path, remote_path, mode='0440', **kwargs):
        self.deploy(local_path, '/tmp/sudoers', mode=mode)
        sudo('mv /tmp/sudoers {0}'.format(remote_path))


def make_deploy(deploy_path, last_segment='deploy'):
    """
    Make a ``deploy()`` function, for uploading files to the server.

    During a deployment, one usually needs to upload certain additional files
    to the server.  It's also often necessary to dynamically define certain
    settings etc. within these files.  The :func:`upload_template()` and
    :func:`put()` functions, respectively, handle uploading files which do and
    do not require dynamic variable substitution.

    The return value from ``make_deploy()`` is a function which will call
    ``put()`` or ``upload_template()`` based on whether or not the file path
    ends with ``'.template'``.

    To make the ``deploy()`` function even simpler for the caller, it will
    assume a certain context for local file paths.  This means one only need
    provide a base file name when calling ``deploy()``, and it will be
    interpreted as relative to the function's context path.

    The ``deploy_path`` argument is used to establish the context path for the
    function.  If it is a folder path, it will be used as-is; otherwise it will
    be constructed by joining the parent folder of ``deploy_path`` with the
    value of ``last_segment``.

    Typical usage then is something like::

       from rattail.fablib import make_deploy

       deploy = make_deploy(__file__)

       deploy('rattail/init-filemon', '/etc/init.d/rattail-filemon',
              mode='0755')

       deploy('rattail/rattail.conf.template', '/etc/rattail.conf')

    This shows what is intended to be typical, i.e. where ``__file__`` is the
    only argument required for ``make_deploy()``.  For the above to work will
    require you to have something like this file structure, where
    ``fabfile.py`` is the script which contains the above code::

       myproject/
       |-- fabfile.py
       |-- deploy/
           `-- rattail/
               |-- init-filemon
               |-- rattail.conf.template
    """
    return Deployer(deploy_path, last_segment)


def rsync(host, *paths):
    """
    Runs rsync as root, for the given host and file paths.
    """
    for path in paths:
        assert path.startswith('/')
        path = path.rstrip('/')
        cmd = 'SSH_AUTH_SOCK=$SSH_AUTH_SOCK rsync -aP --del root@{0}:{1}/ {1}'.format(host, path)
        with settings(forward_agent=True):
            sudo(cmd, shell=False)
